import numpy as np

def update_vdj(adata,vdj_df):

    AB_list = ['TRA','TRB']
    gene_key_list = ['v_gene','d_gene','j_gene']

    c_5GEX = adata.obs['method']=='5GEX'

    tcr_df = vdj_df[vdj_df.chain.isin(AB_list)] # select tcra b gene info
    gene_list = [x for x in set(list(tcr_df.v_gene)+list(tcr_df.j_gene)+list(tcr_df.d_gene)) if x !='None']

    c1 = vdj_df.full_length == True
    c2 = vdj_df.productive == 'True' # False
    c3 = vdj_df.umis > 1
    c4 = vdj_df.productive == 'False'
   
    # fl
    
    is_TRA = np.zeros(len(adata),dtype=bool)
    is_TRB = np.zeros(len(adata),dtype=bool)
    
    for gene in gene_list:
        if ('RBV' in gene) | ('RAV' in gene):
            gene_key = 'v_gene'
        elif ('RBD' in gene) | ('RAD' in gene):
            gene_key = 'd_gene'
        elif ('RBJ' in gene) | ('RAJ' in gene):
            gene_key = 'j_gene'
        else:
            print(gene)
            raise SystemError

        c_gene = vdj_df[gene_key] == gene
        c_vdj = c1&c3&c_gene #c1&c2&c3&c_gene
        vdj_list = list(vdj_df[c_vdj].obs_name)
        c_vdj = adata.obs_names.isin(vdj_list)
        
        if gene.startswith('TRA'):
            is_TRA += c_vdj
        elif gene.startswith('TRB'):
            is_TRB += c_vdj
        else:
            raise SystemError
    
    adata.obs['is_TRA_fl'] = is_TRA
    adata.obs['is_TRB_fl'] = is_TRB 
        
    # np
    
    is_TRA = np.zeros(len(adata),dtype=bool)
    is_TRB = np.zeros(len(adata),dtype=bool)
    
    for gene in gene_list:
        if ('RBV' in gene) | ('RAV' in gene):
            gene_key = 'v_gene'
        elif ('RBD' in gene) | ('RAD' in gene):
            gene_key = 'd_gene'
        elif ('RBJ' in gene) | ('RAJ' in gene):
            gene_key = 'j_gene'
        else:
            print(gene)
            raise SystemError

        c_gene = vdj_df[gene_key] == gene
        c_vdj = c1&c4&c3&c_gene #c1&c2&c3&c_gene
        vdj_list = list(vdj_df[c_vdj].obs_name)
        c_vdj = adata.obs_names.isin(vdj_list)
        
        if gene.startswith('TRA'):
            is_TRA += c_vdj
        elif gene.startswith('TRB'):
            is_TRB += c_vdj
        else:
            raise SystemError
    
    adata.obs['is_TRA_np'] = is_TRA
    adata.obs['is_TRB_np'] = is_TRB
    
    # productive
    
        
    is_TRA = np.zeros(len(adata),dtype=bool)
    is_TRB = np.zeros(len(adata),dtype=bool)
    
    for gene in gene_list:
        if ('RBV' in gene) | ('RAV' in gene):
            gene_key = 'v_gene'
        elif ('RBD' in gene) | ('RAD' in gene):
            gene_key = 'd_gene'
        elif ('RBJ' in gene) | ('RAJ' in gene):
            gene_key = 'j_gene'
        else:
            print(gene)
            raise SystemError

        c_gene = vdj_df[gene_key] == gene
        c_vdj = c1&c2&c3&c_gene #c1&c2&c3&c_gene
        vdj_list = list(vdj_df[c_vdj].obs_name)
        c_vdj = adata.obs_names.isin(vdj_list)
        adata.obs['VDJ_'+gene] = c_vdj
        
        if gene.startswith('TRA'):
            is_TRA += c_vdj
        elif gene.startswith('TRB'):
            is_TRB += c_vdj
        else:
            raise SystemError
    
    adata.obs['is_TRA_p'] = is_TRA
    adata.obs['is_TRB_p'] = is_TRB


from collections import Counter
from collections import defaultdict
import scanpy as sc
import scrublet as scr
import pandas as pd
import pickle as pkl
from .markers import find_markers, show_marker_plot
from .model import generate_training_X
import numpy as np
from bbknn import bbknn

import scipy
import matplotlib.pyplot as plt
import seaborn as sns
import re
import glob
import os
import sys
from geosketch import gs

def create_directory(folder_name):
    if not os.path.exists(folder_name):
        print('no existing folder, making {}...'.format(folder_name))
        print(os.system('mkdir {}'.format(folder_name)))
    else:
        print('{} folder already exists'.format(folder_name))

def calculate_cell_count(adata,anno_key,donor_key = 'donor'):

    meta_table = adata.obs
    donor_list = np.unique(meta_table[donor_key])
    count_dict = {}

    # uncorrected raw counts
    s_list = []
    for donor in donor_list:
        c_donor = (meta_table[donor_key] == donor)
        if np.sum(c_donor)==0:
            continue
        order = str(int(adata.obs['Order'][c_donor][0]))
        age = adata.obs['Age'][c_donor][0]
        stage = adata.obs['Stage'][c_donor][0]

        s_list.append(meta_table[c_donor].groupby(anno_key)['Sample'].count())
        s_list[-1] = s_list[-1].rename("%s_%s_%s_%s"%(order, donor, age, stage))

    s_list = sorted(s_list,key=lambda x: int(x.name.split("_")[0]))
    count_dict['raw_count'] = pd.DataFrame.from_items([(s.name,s) for s in s_list])
    count_dict['raw_ratio'] = count_dict['raw_count']/count_dict['raw_count'].sum(axis=0)

    # corrected counts -> only calculated for CD45PN or CD3PN or TOT samples
    c_excl = ~meta_table['sort'].isin('45NM,EPCAM,CD137,MAIT'.split(","))
    s_list = []
    for donor in donor_list:
        c_donor = (meta_table[donor_key] == donor) & c_excl
        if np.sum(c_donor)==0:
            continue
        order = str(int(adata.obs['Order'][c_donor][0]))
        age = adata.obs['Age'][c_donor][0]
        stage = adata.obs['Stage'][c_donor][0]

        s_list.append(meta_table[c_donor].groupby(anno_key)['factor'].sum())
        s_list[-1] = s_list[-1].rename("%s_%s_%s_%s"%(order, donor, age, stage))


    s_list = sorted(s_list,key=lambda x: int(x.name.split("_")[0]))
    count_dict['norm_count'] = pd.DataFrame.from_items([(s.name,s) for s in s_list])
    count_dict['norm_ratio'] = count_dict['norm_count']/count_dict['norm_count'].sum(axis=0)
    
    return count_dict


def get_df_for_cell_population(cell_list,cnt_df,save_loc=None,figsize=(6,2)):
    
    from scipy.interpolate import make_interp_spline, BSpline
    from statsmodels.nonparametric.smoothers_lowess import lowess
    
    df = cnt_df[cnt_df.index.isin(cell_list)]
    df = (df/df.sum(axis=0))*100
    df = df.reindex(cell_list)

    plt.figure(figsize=figsize)
    ax = plt.subplot(111)

    for ct in cell_list:
        xlen = len(df.columns)
        x1 = np.arange(xlen)
        y1 = np.array(df[df.index==ct])[0]
        mask = ~np.isnan(y1)
        x1 = x1[mask]
        y1 = y1[mask]
        ax.plot(x1,y1,marker='o', 
                alpha=0.5,linewidth=0.5,label=ct)

    ax.grid(False)
    ax.set_xticks(np.arange(xlen))
    ax.set_xticklabels(df.columns)
    for tick in ax.xaxis.get_major_ticks():
            tick.label.set_fontsize(10) 
            #specify integer or one of preset strings, e.g.
            #tick.label.set_fontsize('x-small') 
            tick.label.set_rotation('vertical')
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.legend(bbox_to_anchor=(1.0,0.8))
    if save_loc:
        plt.savefig(save_loc,bbox_inches='tight',format='pdf',dpi=300)
    return df


def get_df_for_cell_population_v2(cell_list,count_dict,style='norm',
                                  save_loc=None,figsize=(6,2),ylim=None,size=20):
    
    from scipy.interpolate import make_interp_spline, BSpline
    from statsmodels.nonparametric.smoothers_lowess import lowess
    
    raw_cnt_df = count_dict['%s_count'%style]
    raw_cnt_df = raw_cnt_df.fillna(0)
    cnt_df = count_dict['%s_ratio'%style]
    
    df = cnt_df[cnt_df.index.isin(cell_list)]
    df = (df/df.sum(axis=0))*100
    df = df.reindex(cell_list)

    plt.figure(figsize=figsize)
    ax = plt.subplot(111)

    counts = np.zeros(len(raw_cnt_df.columns))
    for i, ct in enumerate(cell_list):
        xlen = len(df.columns)
        x1 = np.arange(xlen)
        y1 = np.array(df[df.index==ct])[0]
        cnts = np.array(raw_cnt_df[raw_cnt_df.index==ct])[0]
        mks = np.log10(cnts)*size
        counts+=cnts

        mask = ~np.isnan(y1)
        x1 = x1[mask]
        y1 = y1[mask]
        mks = list(mks[mask])
        
        ax.scatter(x1-0.03*len(cell_list)+0.03*i,y1,s=mks,marker='o',alpha=0.9,label=ct,linewidth=0)  
       # ax.plot(x1,y1,marker='.', markersize=0, 
       #         alpha=0.5,linewidth=0.5,label=ct)


    ax.grid(False)
    ax.set_xticks(np.arange(xlen))
    ax.set_xticklabels([k.split("_")[-2] for k in df.columns])
    #ax.set_xticklabels([' '.join(k.split("_")[-2:]) for k in df.columns])
    for tick in ax.xaxis.get_major_ticks():
            tick.label.set_fontsize(10) 
            #specify integer or one of preset strings, e.g.
            #tick.label.set_fontsize('x-small') 
            tick.label.set_rotation('vertical')
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    #ax.legend(bbox_to_anchor=(1.3,0.8))
    lgnd = ax.legend(bbox_to_anchor=(1.3,0.8), scatterpoints=1, fontsize=10)

    for i in range(len(lgnd.legendHandles)):
        lgnd.legendHandles[i]._sizes = [10]
    
    if ylim:
        plt.ylim(-5,ylim)
    if save_loc:
        plt.savefig(save_loc,bbox_inches='tight',format='pdf',dpi=300)
    return df,counts

def csr_vappend(a,b):
    """ Takes in 2 csr_matrices and appends the second one to the bottom of the first one. 
    Much faster than scipy.sparse.vstack but assumes the type to be csr and overwrites
    the first matrix instead of copying it. The data, indices, and indptr still get copied."""

    a.data = np.hstack((a.data,b.data))
    a.indices = np.hstack((a.indices,b.indices))
    a.indptr = np.hstack((a.indptr,(b.indptr + a.nnz)[1:]))
    a._shape = (a.shape[0]+b.shape[0],b.shape[1])
    return a

def ravel_index(pos, shape):
    ''' inverse of np.unravel_index'''
    res = 0
    acc = 1
    for pi, si in zip(reversed(pos), reversed(shape)):
        res += pi * acc
        acc *= si
    return res

def flatten(l):
    return [item for sublist in l for item in sublist]

def intersect(*d):
    sets = list(map(set, d))
    result = sets[0]
    for s in sets[1:]:
        result = result.intersection(s)
    return result


from collections import Counter
from collections import defaultdict
import scanpy as sc
import pandas as pd
import pickle as pkl
from .genes import cc_genes
from .colors import vega_20, vega_20_scanpy, zeileis_26, godsnot_64
import numpy as np
from bbknn import bbknn
import scipy
import matplotlib.pyplot as plt
import seaborn as sns


### network derivation

# select = get_grid(bdata,n_neighbor=n_neighbor,select_per_grid = 20,scale=1)
# idata = impute_neighbor(bdata,n_neighbor=n_neighbor)
# subset = ~bdata.obs['anno_v24'].isin(["NMP"])
# select_subset = [x for x in select if x in np.where(subset)[0]]
# tfdata = new_exp_matrix(bdata,idata,select_subset,tflist=tf_lists,max_cutoff=0.1,ratio_expressed=0.005,min_disp=0.5,example_gene='CREB3L3')
# generate_gene_network(tfdata,n_neighbors=n_neighbor)
# anno_key = 'anno_v24'
# anno_uniq, anno_ratio = impute_anno(bdata,select_subset,anno_key,n_neighbor=n_neighbor)
# draw_graph(tfdata, anno_key, anno_uniq, anno_ratio,adjust=True)


class network():
    
    def __init__(self,adata,n_neighbor=10):
        print('calculating 3d umap')
        
        self.bdata = adata.copy()
        sc.tl.umap(self.bdata,n_components = 3)
        self.n_neighbor = 10
        self.anno_dict = {}
    
    def get_grid(self,scale=2,border=2,select_per_grid=20):
        self.select = get_grid(self.bdata,n_neighbor=self.n_neighbor,
                               select_per_grid = select_per_grid,scale=scale,border = border)
    def impute(self):
        self.idata = impute_neighbor(self.bdata,n_neighbor=self.n_neighbor)
    
    def process_genes(self, gene_set = None, C_cut = 0, **kwargs):
        '''run imput_neighbor, new_exp_matrix, calculate correlation and generate network
        :::new_exp_matrix args 
        max_cutoff = 0.2,ratio_expressed = 0.01,
        min_disp = .00, n_min_exp_cell = 50, example_gene='EPCAM'''
        self.tfdata = new_exp_matrix(self.bdata,self.idata,self.select,
                                             tflist = gene_set, **kwargs)
        
        print(self.tfdata.shape)
        
        C = scipy.sparse.csr_matrix(np.corrcoef(self.tfdata.X.todense()))
        plt.imshow(C.todense())
        plt.show()
        C[C<C_cut] = 0

        self.tfdata.uns['neighbors'] = {}
        self.tfdata.uns['neighbors']['connectivities'] = C
        self.tfdata.uns['neighbors']['params'] = {'n_neighbors': self.n_neighbor, 'method': 'umap', 'metric': 'cosine'}
        
        generate_gene_network(self.tfdata)
        
    def annotation(self, anno_key):
        self.anno_dict[anno_key] = impute_anno(self.bdata,self.select,anno_key, n_neighbor = self.n_neighbor)
    
    
    def draw_network(self, anno_key, **kwargs):
        '''params:::
        factor0=100.0, adjust=False,
        text_fontsize=8,z_score_cut=0.4,
        color_dict=None, axis='X_draw_graph_kk'''
        anno_uniq = self.anno_dict[anno_key][0]
        anno_ratio = self.anno_dict[anno_key][1]
        draw_graph(self.tfdata, anno_key, anno_uniq, anno_ratio, **kwargs)
        
        
        
        




def get_grid(bdata, scale=1, border=2, expand = 3,
           select_per_grid=5, min_count = 2, n_neighbor = 10):
    
    import math
    from collections import defaultdict
    def shift(crd, pos):
        if (min(crd[:,pos])<0): crd[:,pos]+=abs(min(crd[:,pos]))
        if (max(crd[:,pos])<0): crd[:,pos]+=abs(max(crd[:,pos]))
    
    # get umap coordinate
    crd = bdata.obsm['X_umap'] 
    
    shift(crd,0)
    shift(crd,1)
    shift(crd,2)
    
    picture = np.zeros((scale*math.ceil(max(crd[:,0])-min(crd[:,0]))+border*scale*expand, 
                        scale*math.ceil(max(crd[:,1])-min(crd[:,1]))+border*scale*expand,
                        scale*math.ceil(max(crd[:,2])-min(crd[:,2]))+border*scale*expand))

    for pos in crd*scale+border*scale/2:
        picture[math.floor(pos[0]),math.floor(pos[1]),math.floor(pos[2])]+=1
    
    plt.imshow(np.sum(picture,axis=1),vmin=10)
    plt.grid(False)
    plt.show()
    
    plt.hist(picture[picture>5],bins=100)
    plt.show()
    
    # prepare grid
    grid = defaultdict(list)
    for idx, pos in enumerate(crd*scale+border*scale/2):
        posid = '%i:%i:%i'%(math.floor(pos[0]),math.floor(pos[1]),math.floor(pos[2]))
        grid[posid].append(idx)
    
    # select grid which has more than [min_count] cells and np.min(grid_size,select_per_grid) number of representative cells from grid
    np.random.seed(0)

    select = []
    for posid in grid:
        grid_size = len(grid[posid])
        if grid_size < min_count:
            continue
        else:
            select.extend(np.random.choice(grid[posid],size=min([grid_size,select_per_grid]),replace=False))
    
    return select

def impute_neighbor(bdata,n_neighbor=10):
    from scipy.spatial import cKDTree
    from sklearn.neighbors import KDTree
    import multiprocessing as mp

    n_jobs = mp.cpu_count()
    
    # Get neighborhood structure based on 
    ckd = cKDTree(bdata.obsm["X_umap"])
    ckdout = ckd.query(x=bdata.obsm["X_umap"], k=n_neighbor, n_jobs=n_jobs)
    indices = ckdout[1]
    
    sum_list = []
    import scipy
    for i in range(0,bdata.raw.X.shape[0],10000):
        start = i
        end = min(i+10000,bdata.raw.X.shape[0])
        X_list = [bdata.raw.X[indices[start:end,i]] for i in range(n_neighbor)]
        X_sum = scipy.sparse.csr_matrix(np.sum(X_list)/n_neighbor)
        sum_list.append(X_sum)
        print(i)
        
    imputed = scipy.sparse.vstack(sum_list)
    idata = sc.AnnData(imputed)
    idata.obs = bdata.obs.copy()
    idata.var = bdata.raw.var.copy()
    idata.obsm = bdata.obsm.copy()
    idata.uns = bdata.uns.copy()
    
    return idata

def new_exp_matrix(bdata,idata,select,n_min_exp_cell = 10, min_mean=0,min_disp=.1, ratio_expressed = 0.1,example_gene='CDK1',show_filter = None, max_cutoff=0.2, tflist = None, calc_var = False):
    
    # get genes expressed more than min_exp_cell
    detected = np.sum(bdata.raw.X>0,axis=0).A1
    
    Xnew = idata.X[select].todense()
    
    import scipy
    gdata = sc.AnnData(scipy.sparse.csr_matrix(Xnew))
    gdata.var_names = bdata.raw.var_names
    gdata.raw = sc.AnnData(Xnew)
    
    if calc_var == True: # otherwise, skip highly variable gene selection
    
        # select highly variable genes
        print('selecting hvgs...')
        result = sc.pp.filter_genes_dispersion(gdata.X,log=False,min_mean=min_mean,min_disp=min_disp)

        if example_gene:
            pos = np.where(gdata.var_names==example_gene)[0][0]
            plt.hist(gdata.X[:,pos].todense().A1)
            plt.show()
            print('max:',np.max(gdata.X[:,pos]))
            print('mean:',np.mean(gdata.X[:,pos]))
            print('min:',np.min(gdata.X[:,pos]))
            print('dispersions_norm:',result['dispersions_norm'][pos])

        if show_filter:
            x = result.means
            y = result.dispersions_norm
            c = result.gene_subset
            print(np.sum(c), 'highly variable genes are selected')
            plt.scatter(x,y)
            plt.scatter(x[c],y[c])
            plt.show()

        gene_info = gdata.copy()
        filter_result = result.copy()

        # do filter
        c1 = (result.gene_subset) # highly variable above min_disp
        c2 = (np.max(gdata.X,axis=0).todense().A1 > max_cutoff) # max expression should be above max_cutoff
        c3 = np.sum(gdata.X>max_cutoff,axis=0).A1 > ratio_expressed*len(gdata.obs_names)
        c4 = detected > n_min_exp_cell
        deg = (c1 & c3 & c4)
        gdata = gdata[:,deg].copy()
    
    else:
        pass
    
    # invert gene to cell
    import scipy
    cdata = sc.AnnData(scipy.sparse.csr_matrix(gdata.X.T))
    cdata.obs_names = gdata.var_names
    
    if tflist:
        tf_idx = cdata.obs_names.isin(tflist)
        tfdata = cdata[tf_idx].copy()
        return tfdata #, gene_info, filter_result
    else:
        return cdata


def generate_gene_network(tfdata):
    #sc.pp.pca(tfdata)
    #sc.pp.neighbors(tfdata,metric='cosine',n_neighbors=n_neighbors)
    sc.tl.umap(tfdata,min_dist=0.7)
    sc.tl.draw_graph(tfdata,layout='fa')
    sc.tl.draw_graph(tfdata,layout='fr')
    sc.tl.draw_graph(tfdata,layout='kk')
    
def impute_anno(bdata,select, anno_key,n_neighbor=10):
    from scipy.spatial import cKDTree
    from sklearn.neighbors import KDTree
    import multiprocessing as mp

    n_jobs = mp.cpu_count()

    # Get neighborhood structure based on 
    ckd = cKDTree(bdata.obsm["X_umap"])
    ckdout = ckd.query(x=bdata.obsm["X_umap"], k=n_neighbor, n_jobs=n_jobs)
    indices = ckdout[1]
    anno_uniq = sorted(set(bdata.obs[anno_key]))
    anno_arr = np.vstack([np.array(bdata.obs[anno_key]==x).astype(int) for x in anno_uniq])
    anno_sum = np.zeros(shape=anno_arr.shape)

    for i in range(n_neighbor):
        anno_sum += anno_arr[:,indices[:,i]]

    anno_ratio = anno_sum/np.sum(anno_sum,axis=0)
    select_anno = [x in set(bdata.obs[anno_key][select]) for x in anno_uniq]
    
    return np.array(anno_uniq)[select_anno], anno_ratio[select_anno][:,select]


def draw_graph(tfdata, anno_key, anno_uniq, anno_ratio, factor0=100.0, adjust=False,
           text_fontsize=8,z_score_cut=0.4,color_dict=None, axis='X_draw_graph_kk'):
    import seaborn as sns
    palette1 =    ["#023fa5", "#7d87b9", "#bec1d4", "#d6bcc0", "#bb7784", "#8e063b", "#4a6fe3",
        "#8595e1", "#b5bbe3", "#e6afb9", "#e07b91", "#d33f6a", "#11c638", "#8dd593",
        "#c6dec7", "#ead3c6", "#f0b98d", "#ef9708", "#0fcfc0", "#9cded6", "#d5eae7",
        "#f3e1eb", "#f6c4e1", "#f79cd4",
        '#7f7f7f', "#c7c7c7", "#1CE6FF", "#336600"]

    palette2 = [
        '#1f77b4', '#ff7f0e', '#2ca02c', '#d62728',
        '#9467bd', '#8c564b', '#e377c2',  # '#7f7f7f' removed grey
        '#bcbd22', '#17becf',
        '#aec7e8', '#ffbb78', '#98df8a', '#ff9896',
        '#c5b0d5', '#c49c94', '#f7b6d2',  # '#c7c7c7' removed grey
        '#dbdb8d', '#9edae5',
        '#ad494a', '#8c6d31']  # manual additions

    
    if len(anno_uniq)> 28:
        palette = godsnot_64
    elif len(anno_uniq)> 20:
        palette = palette1
    else:
        palette = palette2
        
    if color_dict:
        palette = [color_dict[x] for x in anno_uniq]
        
    Cr = np.corrcoef(np.vstack([tfdata.X.todense(),anno_ratio]))
    Cr_an = Cr[len(tfdata):,:len(tfdata)]
    color_anno = np.argmax(Cr_an,axis=0)
    
    C = tfdata.uns['neighbors']['connectivities'].todense()
    #C = Cr[:len(tfdata),:len(tfdata)]
    pairs = np.where(C>0.3)

    from sklearn import preprocessing

    Cr_scaled = preprocessing.scale(Cr_an)
    Cr_scaled = Cr_an
    #dot_size0 = np.choose(color_anno,Cr_scaled)
    dot_size0 = np.array([Cr_scaled[j,i] for i,j in enumerate(color_anno)])
    
    colors = np.array([palette[i] if s >z_score_cut else 'gray' for i,s in zip(color_anno,dot_size0)])
    
    from adjustText import adjust_text

    show = True
    gamma = 2
    dot_size = dot_size0**(gamma)

    x = tfdata.obsm[axis][:,0]
    y = tfdata.obsm[axis][:,1]
    n = list(tfdata.obs_names)

    plt.figure(figsize=(8,8))
    plt.scatter(x,y,c=colors,s=factor0*dot_size) #50*size

    # draw pair to pair lines
    for p1, p2 in zip(pairs[0],pairs[1]):
        plt.plot([x[p1],x[p2]],[y[p1],y[p2]],c='lightgrey',alpha=0.3,zorder=-1,
                linewidth=2*C[p1,p2]**gamma)

    # draw_label
    for i, label in enumerate(anno_uniq):
        plt.scatter(0,0,s=0,label=label,
                    c=palette[i],zorder=1,alpha=1.0,
                    linewidth=0)
    # add_names
    if show != False:   
        texts = []
        for i, txt in enumerate(n):
            if txt in anno_uniq:
                texts.append(plt.text(x[i],y[i],txt,fontsize=text_fontsize*1.2,color=color_dict[txt]))
            else:
                texts.append(plt.text(x[i],y[i],txt,fontsize=text_fontsize))

    lgnd = plt.legend(loc=(1.1,0), scatterpoints=1, fontsize=10)
    for handle in lgnd.legendHandles:
        handle.set_sizes([30.0])
        handle.set_alpha(1) 

    plt.xticks([], [])
    plt.yticks([], [])
    plt.grid(False)
    if adjust == True:
        adjust_text(texts,only_move={'text':'y'})


import numpy as np
from collections import Counter
from collections import defaultdict
import matplotlib.pyplot as plt
import plotly
import plotly.plotly as py
import plotly.graph_objs as go

def draw_3d(adata,group_key,version,size = 2):

    color_list = adata.uns[group_key+'_colors']
    trace_list = []

    for i, ct in enumerate(set(adata.obs[group_key])):
        sdata = adata[adata.obs[group_key]==ct]

        x = sdata.obsm['X_umap'][:,0]
        y = sdata.obsm['X_umap'][:,1]
        z = sdata.obsm['X_umap'][:,2]

        trace = go.Scatter3d(
            x=x,
            y=y,
            z=z,
            name=ct,
            mode='markers',
            marker=dict(
                size=size,
                color=color_list[i],
                opacity=0.8
            )
        )

        trace_list.append(trace)

    data = trace_list
    layout = go.Layout(
        margin=dict(
            l=0,
            r=0,
            b=0,
            t=0
        )
    )

    fig = go.Figure(data=data, layout=layout)
    plotly.offline.plot(fig, filename='%s-%s-umap-3d-scatter.html'%(version,group_key))
    
def draw_dotplot_scale(fig_folder):
    figscale= 1.5
    fig = plt.figure(figsize=(3,2))
    ax = fig.add_subplot(1,1,1)
    xpos = [1,1.27,1.61,2]
    x1 = np.array([0.25,0.5,0.75,1])
    ypos = [1,1,1,1]
    ax.scatter(xpos,ypos,s=x1*figscale*400,c='grey',cmap='OrRd',marker='o')
    ax.set_xlim(0.5,2.5)
    plt.grid(False)
    plt.savefig(fig_folder+'Dopplot.scale.pdf',format='pdf')
    
    
def draw_pseudo_heatmap(adata,genelist):
    
    import sklearn
    def running_mean(x, N):
        cumsum = np.cumsum(np.insert(x, 0, 0)) 
        return (cumsum[N:] - cumsum[:-N]) / float(N)
    
    Xt = adata.raw.X
    Rt = Xt.expm1()
    t = adata.obs['dpt_order']
    Rts = Rt[np.argsort(np.array(t))]
    bins = np.linspace(0,len(t),20)
    binlabel = np.digitize(np.array(t), bins)
    adata.obs['dpt_bins'] = [str(x) for x in binlabel]

    gidx = adata.raw.var_names.isin(genelist)
    gene_order = list(adata.raw.var_names[gidx])
    re_order = [gene_order.index(x) for x in genelist]

    normed = sklearn.preprocessing.normalize(Rts[:,gidx][:,re_order].todense().T,norm='max')
    runned = np.apply_along_axis(lambda x: running_mean(x,100),1,normed)
    normed = sklearn.preprocessing.normalize(runned,norm='max')

    fig = plt.figure(figsize=(3,3))
    ax = fig.add_subplot(111)

    height=12
    ax.imshow(normed, extent = [0,10,0,height], vmax=1,origin='upper',cmap='RdYlBu_r')
    ax.set_yticks(np.linspace(0,height,len(genelist)+1)[:-1]+height/(len(genelist)*2))
    ax.set_yticklabels(genelist[::-1],fontsize=6)
    ax.set_xticks([])
    ax.grid(False)
    #ax.set_aspect(0.1)
    plt.show()

def draw_pseudo_heatmap_anno(adata,anno,color_list =None,anno_list=None):
    import sklearn
    def running_mean(x, N):
        cumsum = np.cumsum(np.insert(x, 0, 0)) 
        return (cumsum[N:] - cumsum[:-N]) / float(N)

    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib.colors import LinearSegmentedColormap
    from matplotlib import colors

    t = adata.obs['dpt_order']
    anl = list(set(adata.obs[anno]))
    Ob = np.vstack([np.array(adata.obs[anno]==an) for an in anl]).astype(int).T
    Obs = Ob[np.argsort(np.array(t))]
    
    order = anl
    if anno_list!=None:
        re_order = [anl.index(x) for x in anno_list]
    else:
        re_order = [anl.index(x) for x in anl]
        anno_list = anl

    normed = sklearn.preprocessing.normalize(Obs[:,re_order].T,norm='max')
    runned = np.apply_along_axis(lambda x: running_mean(x,1),1,normed)
    normed = sklearn.preprocessing.normalize(runned,norm='max')

    fig = plt.figure(figsize=(3,0.5))
    ax = fig.add_subplot(111)

    height=1.4

    clist = ['white', 'red','green','blue','orange','skyblue','cyan','magenta']
    clist = ['white']+adata.uns[anno+'_colors']
    if clist:
        clist = ['white']+color_list

    for i in range(len(anl)):
        print(i, height/len(anl)*i,height/len(anl)*(i+1), anno_list[i])
        cmap = colors.ListedColormap([clist[0],clist[i+1]])
        ax.imshow(normed[i:i+1], extent = [0,10,height/(len(anl))*i,height/(len(anl))*(i+1)], vmax=1,origin='upper',cmap=cmap)
    ax.set_yticks(np.linspace(0,height,len(anno_list)+1)[:-1]+height/(len(anno_list)*2))
    ax.set_yticklabels(anno_list,fontsize=8)
    ax.set_xticks([])
    ax.set_ylim(0,height)
    ax.grid(False)
    #ax.set_aspect(0.1)
    plt.show()
    

def draw_pseudo_heatmap_v2(adata,genelist,anno=None, clist=None, anno_list =None, gridspec_kw = {'height_ratios': [10,1]},figsize=(3,4),fontsize=12, mean_window = 100, anno_window=10):
    
    import sklearn
    def running_mean(x, N):
        cumsum = np.cumsum(np.insert(x, 0, 0)) 
        return (cumsum[N:] - cumsum[:-N]) / float(N)
    
    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib.colors import LinearSegmentedColormap
    from matplotlib import colors
    
    
    fig,(axes) = plt.subplots(2, 1,  sharex= True,
                        gridspec_kw=gridspec_kw,
                        figsize=figsize)
    
    
    Xt = adata.raw.X
    Rt = Xt.expm1()
    t = adata.obs['dpt_order']
    Rts = Rt[np.argsort(np.array(t))]
    bins = np.linspace(0,len(t),20)
    binlabel = np.digitize(np.array(t), bins)
    adata.obs['dpt_bins'] = [str(x) for x in binlabel]

    gidx = adata.raw.var_names.isin(genelist)
    gene_order = list(adata.raw.var_names[gidx])
    re_order = [gene_order.index(x) for x in genelist]

    normed = sklearn.preprocessing.normalize(Rts[:,gidx][:,re_order].todense().T,norm='max')
    runned = np.apply_along_axis(lambda x: running_mean(x,mean_window),1,normed)
    normed = sklearn.preprocessing.normalize(runned,norm='max')
    

    ax = axes[0]

    height=12
    im = ax.imshow(normed, extent = [0,10,0,height], vmax=1,origin='upper',cmap='RdYlBu_r')
    ax.set_yticks(np.linspace(0,height,len(genelist)+1)[:-1]+height/(len(genelist)*2))
    ax.set_yticklabels(genelist[::-1],fontsize=fontsize)
    ax.set_xticks([])
    ax.grid(False)
    ax.set_aspect('auto')
    
    if anno:

        anl = sorted(list(set(adata.obs[anno])))
        Ob = np.vstack([np.array(adata.obs[anno]==an) for an in anl]).astype(int).T
        Obs = Ob[np.argsort(np.array(t))]

        order = anl
        if anno_list!=None:
            re_order = [anl.index(x) for x in anno_list]
        else:
            re_order = [anl.index(x) for x in anl]
            anno_list = anl

        normed = sklearn.preprocessing.normalize(Obs[:,re_order].T,norm='max')
        runned = np.apply_along_axis(lambda x: running_mean(x,anno_window),1,normed)
        normed = sklearn.preprocessing.normalize(runned,norm='max')
        normed = normed>0
        ax = axes[1]
        
        height=1.4


        if clist:
            clist = ['white']+clist
        else:
            clist = ['white']+list(np.array(adata.uns[anno+"_colors"])[re_order])

        for i in range(len(anl)):
            print(i, height/len(anl)*i,height/len(anl)*(i+1), anno_list[i])
            cmap = colors.ListedColormap([clist[0],clist[i+1]])
            ax.imshow(normed[i:i+1], extent = [0,10,height/(len(anl))*i,height/(len(anl))*(i+1)], vmax=1,origin='upper',cmap=cmap)
        ax.set_yticks(np.linspace(0,height,len(anno_list)+1)[:-1]+height/(len(anno_list)*2))
        ax.set_yticklabels(anno_list,fontsize=fontsize)
        ax.set_xticks([])
        ax.set_ylim(0,height)
        ax.grid(False)
        ax.set_aspect('auto')

    plt.subplots_adjust(wspace=0, hspace=0.01)
    
    return im

# copied from https://gist.github.com/wflynny/79c5266cc39a4a884958d696f84f85df

import scanpy.api as sc
import matplotlib.pyplot as plt
import seaborn as sns

def get_cluster_proportions(adata,
                            cluster_key="cluster_final",
                            sample_key="replicate",
                            drop_values=None):
    """
    Input
    =====
    adata : AnnData object
    cluster_key : key of `adata.obs` storing cluster info
    sample_key : key of `adata.obs` storing sample/replicate info
    drop_values : list/iterable of possible values of `sample_key` that you don't want
    
    Returns
    =======
    pd.DataFrame with samples as the index and clusters as the columns and 0-100 floats
    as values
    """
    
    adata_tmp = adata.copy()
    sizes = adata_tmp.obs.groupby([cluster_key, sample_key]).size()
    props = sizes.groupby(level=1).apply(lambda x: 100 * x / x.sum()).reset_index() 
    props = props.pivot(columns=sample_key, index=cluster_key).T
    props.index = props.index.droplevel(0)
    props.fillna(0, inplace=True)
    
    if drop_values is not None:
        for drop_value in drop_values:
            props.drop(drop_value, axis=0, inplace=True)
    return props


def plot_cluster_proportions(cluster_props, 
                             cluster_palette=None,
                             xlabel_rotation=0): 
    fig, ax = plt.subplots(dpi=300)
    fig.patch.set_facecolor("white")
    
    cmap = None
    if cluster_palette is not None:
        cmap = sns.palettes.blend_palette(
            cluster_palette, 
            n_colors=len(cluster_palette), 
            as_cmap=True)
   
    cluster_props.plot(
        kind="bar", 
        stacked=True, 
        ax=ax, 
        legend=None, 
        colormap=cmap
    )
    
    ax.legend(bbox_to_anchor=(1.01, 1), frameon=False, title="Cluster")
    sns.despine(fig, ax)
    ax.tick_params(axis="x", rotation=xlabel_rotation)
    ax.set_xlabel(cluster_props.index.name.capitalize())
    ax.set_ylabel("Proportion")
    fig.tight_layout()
    
    return fig


import numpy as np
from collections import Counter
from collections import defaultdict
import matplotlib.pyplot as plt


class marker():
    
    def __init__(self,adata,label,**kwargs):
        self.adata = adata
        self.label = label
        self.mks = find_markers(adata,label,**kwargs)
    
    def plot_marker(self,**kwargs):
        return show_marker_plot(self.adata,self.label,self.mks,**kwargs)
    
    def show_marker(self,celltype,**kwargs):
        return show_marker(self.mks,celltype,**kwargs)

# Marker gene analysis

def find_markers(adata,groupby,scanpy=None,single=True,thres=0.2,min_mean_cut=0.2, min_drop_cut=0.2,min_cnt_cut=10):
    adata.uns['cdm_'+groupby]=calculate_markers(adata,groupby) 
    if single==False:
        mks=find_markers_multiple(adata,adata.uns['cdm_'+groupby],
                                  thres=thres,min_mean_cut=min_mean_cut, 
                                  min_drop_cut=min_drop_cut, min_cnt_cut=min_cnt_cut)
    else:
        mks=find_markers_single(adata,adata.uns['cdm_'+groupby],
                                thres=thres,min_mean_cut=min_mean_cut, 
                                min_drop_cut=min_drop_cut, min_cnt_cut=min_cnt_cut)
    if scanpy:
        sc.tl.rank_genes_groups(adata,groupby)
        adata.uns['rnk']=pd.DataFrame(adata.uns['rank_genes_groups']['names'])
    return mks

def calculate_markers(adata,group_key,exclude = None, cnt_cut=0, tot_cnt_cut=0):

    # group_key = 'louvain'
    # tot_cnt_cut = minimum cell number for ct, default 0, no cut here
    # cnt_cut = minumum cell number of expressing that gene per cell, otherwise, 0

    ctlist = np.array(sorted(list(set(adata.obs[group_key]))))
    
    if exclude:
        ctlist = [x for x in ctlist if x not in exclude]
    
    mtx = adata.raw
    len_gene = mtx.X.shape[1]

    mean_dt = {}
    cnt_dt = {}
    drop_dt = {}

    for ct in ctlist:
        ct_cond = np.array(adata.obs[group_key]==ct)
        if np.sum(ct_cond)<tot_cnt_cut:
            continue
        #print(ct)
        len_ct = np.sum(ct_cond)
        mtx_ct = mtx.X[ct_cond]
        pos_count = Counter(mtx_ct.nonzero()[1])
        cnt_arr = np.zeros(len_gene)
        for pos in pos_count:
            cnt_arr[pos] = pos_count[pos]
        cnt_dt[ct] = cnt_arr
        cnt_mask = (cnt_arr<cnt_cut)
        drop_dt[ct] = cnt_arr/len_ct
        mean_arr = np.mean(mtx_ct,axis=0).A1
        mean_arr[cnt_mask] = 0
        mean_dt[ct] = mean_arr
    
    cdm_out = {"cnt":cnt_dt, "drop":drop_dt, "mean":mean_dt}
    return cdm_out


def find_markers_groups(adata, cdm_out, groups, thres = 0.2, min_exp_cut=0.1, min_cnt_cut = 5):
    # Select marker gene (multiple)
     
    cnt_dt = cdm_out["cnt"]
    drop_dt = cdm_out["drop"]
    mean_dt = cdm_out["mean"]
    
    ctlist = np.array(sorted(list(drop_dt.keys())))
    marker_group = []
    thres = thres

    for k,gname in enumerate(adata.raw.var_names):  
        x1 = []
        y1 = []
        ct_selected = []
        for ct in ctlist:
            if cnt_dt[ct][k] < min_cnt_cut:
                continue
            x1.append(drop_dt[ct][k]) # get expratio
            y1.append(mean_dt[ct][k]) # get expmean
            ct_selected.append(ct)
        x1 = np.array(x1)
        y1 = np.array(y1)
        if len(y1) == 0:
            continue
        y1 = y1/np.max(y1)
        # sort expratio, expmean, celltypelist
        
        is_group = np.array([x in (groups) for x in ct_selected])
        try: max_y1 = np.max(y1[~is_group])
        except: max_y1 = 0
        try: min_y1 = np.min(y1[is_group])
        except: min_y1 = 0
        
        dff = min_y1-max_y1
        cond = dff>thres
            
        if cond:
            chgpos = (np.where(cond)[0][0])+1
            out = 1
            for i,ct_high in enumerate(list(groups)):
                if mean_dt[ct_high][k] < min_exp_cut:
                    out = 0
            if out == 1:
                marker_group.append((gname,dff)) 
    return  marker_group


def find_markers_multiple(adata, cdm_out, thres = 0.2, min_mean_cut=0.2, min_drop_cut=0.2, min_cnt_cut = 0):
    # Select marker gene (multiple)
    
    cnt_dt = cdm_out["cnt"]
    drop_dt = cdm_out["drop"]
    mean_dt = cdm_out["mean"]
    
    ctlist = np.array(sorted(list(drop_dt.keys())))
    marker_multiple = defaultdict(list)
    thres = thres

    for k,gname in enumerate(adata.raw.var_names):  
        x1 = []
        y1 = []
        ct_selected = []
        for ct in ctlist:
            if cnt_dt[ct][k] < min_cnt_cut:
                continue
            x1.append(drop_dt[ct][k]) # get expratio
            y1.append(mean_dt[ct][k]) # get expmean
            ct_selected.append(ct)
            
        x1 = np.array(x1)
        y1 = np.array(y1)
        if len(y1) == 0:
            continue
        y1 = y1/np.max(y1)
        # sort expratio, expmean, celltypelist
        idx = np.argsort(y1)
        x1 = x1[idx]
        y1 = y1[idx]
        ctlist_sorted = np.array(ct_selected)[idx]
        
        dff = (np.diff(y1))
        cond = dff>thres

        if any(cond):
            chgpos = (np.where(cond)[0][0])
            ct_high = ctlist_sorted[-1]
            if mean_dt[ct_high][k] < min_mean_cut:
                pass
            elif drop_dt[ct_high][k] < min_drop_cut:
                pass
            else:
                marker_multiple[ct_high].append((gname,dff[chgpos]))   
    return  marker_multiple

def find_markers_single(adata, cdm_out, thres = 0.2, min_mean_cut=0.2, min_drop_cut=0.2, min_cnt_cut = 0):
    # Select marker gene (single)
  
    cnt_dt = cdm_out["cnt"]
    drop_dt = cdm_out["drop"]
    mean_dt = cdm_out["mean"]
    
    ctlist = np.array(sorted(list(drop_dt.keys())))
    marker_single = defaultdict(list)

    for k,gname in enumerate(adata.raw.var_names):  
        x1 = []
        y1 = []
        ct_selected = []
        cond = False
        for ct in ctlist:
            if cnt_dt[ct][k] < min_cnt_cut:
                continue
            x1.append(drop_dt[ct][k]) # get expratio
            y1.append(mean_dt[ct][k]) # get expmean
            ct_selected.append(ct)
        if len(y1)==0:
            continue
        else:
            y1 = y1/np.max(y1)
            if len(y1)==1:
                if y1[0]>thres:
                    if mean_dt[ct_selected[0]][k] < min_mean_cut:
                        pass
                    elif drop_dt[ct_selected[0]][k] < min_drop_cut:
                        pass
                    else:
                        marker_single[ct_selected[0]].append((gname,y1[0]))
                else:
                    pass
            else:
                x1 = np.array(x1)
                y1 = np.array(y1)
                # sort expratio, expmean, celltypelist
                idx = np.argsort(y1)
                x1 = x1[idx]
                y1 = y1[idx]    
                ctlist_sorted = np.array(ct_selected)[idx]

                dff = (np.diff(y1))
                cond = dff[-1]>thres

                if cond==True:
                    ct_high = ctlist_sorted[-1] 
                    if mean_dt[ct_high][k] < min_mean_cut:
                        pass
                    elif drop_dt[ct_high][k] < min_drop_cut:
                        pass
                    else:
                        marker_single[ct_high].append((gname,dff[-1]))
    return marker_single 

def find_markers_negative(adata, cdm_out, ctlist = None, thres = 0.2, min_exp_cut=0.1):
    # Select marker gene (single)
  
    cnt_dt = cdm_out["cnt"]
    drop_dt = cdm_out["drop"]
    mean_dt = cdm_out["mean"]
    
    if ctlist == None:
        ctlist = np.array(sorted(list(drop_dt.keys())))
        
    marker_single = defaultdict(list)
    thres = thres

    for k,gname in enumerate(adata.raw.var_names):  
        x1 = []
        y1 = []
        ct_selected = []
        cond = False
        max_value = 0.0
        for ct in ctlist:
            max_value = np.max([max_value, mean_dt[ct][k]])
            x1.append(drop_dt[ct][k]) # get expratio
            y1.append(mean_dt[ct][k]) # get expmean
            ct_selected.append(ct)
        if max_value < min_exp_cut:
            continue
        else:
            y1 = y1/np.max(y1)
            x1 = np.array(x1)
            y1 = np.array(y1)
            # sort expratio, expmean, celltypelist
            idx = np.argsort(y1)
            x1 = x1[idx]
            y1 = y1[idx]
            ctlist_sorted = np.array(ct_selected)[idx]

            dff = (np.diff(y1))
            if gname == 'ENSG00000168685':
                print(ctlist_sorted)
                print(y1)
                print(dff)
            cond = dff[0]>thres

            if cond==True:
                ct_high = ctlist_sorted[0] 
                marker_single[ct_high].append((gname,dff[0]))
    for ct in list(marker_single.keys()):
        marker_single[ct] = sorted([(geneID,drop) for geneID,drop in marker_single[ct]],key=lambda x:-x[1])
    return marker_single 


def show_marker(marker_output, celltype=None, toshow=40, result=False):
    if celltype == None:
        if result == True:
            return sorted([(geneID,drop) for geneID,drop in marker_output],key=lambda x:-x[1])[0:toshow]
        else:
            print(sorted([(geneID,drop) for geneID,drop in marker_output],key=lambda x:-x[1])[0:toshow])
    else:
        if result == True:
            return sorted([(geneID,drop) for geneID,drop in marker_output[celltype]],key=lambda x:-x[1])[0:toshow]
        else:
            print(sorted([(geneID,drop) for geneID,drop in marker_output[celltype]],key=lambda x:-x[1])[0:toshow])

def show_marker_plot(adata,anno_key,mks,toshow=10,T=False):
    mklist = []
    ct_list = sorted(adata.uns['cdm_'+anno_key]['cnt'].keys())
    #ct_list = sorted(ct_list,key=lambda x:int(x))
    report_mks = {}
    
    for ct in ct_list:
        
        try:
            ct_mks = [x[0] for x in show_marker(mks,celltype=ct,result=True)][0:toshow]
        except:
            ct_mks = [' ']
        mklist.extend(ct_mks)
        mklist.append(' ')
        report_mks[ct] = ct_mks
        
    if T:
        draw_marker_blob_T(adata,adata.uns['cdm_'+anno_key],mklist,ctlist=ct_list)
    else:
        draw_marker_blob(adata,adata.uns['cdm_'+anno_key],mklist,ctlist=ct_list,min_drop_ratio=0.0)
    return(report_mks)


def draw_marker(adata, gene, cdm_out, min_cnt_cut = 0):
    

    cnt_dt = cdm_out["cnt"]
    drop_dt = cdm_out["drop"]
    mean_dt = cdm_out["mean"]
    
    ctlist = np.array(sorted(list(drop_dt.keys())))
    k = np.where(adata.raw.var_names == gene)[0][0]
    x1 = []
    y1 = []
    ct_selected = []
    for ct in ctlist:
        if cnt_dt[ct][k] < min_cnt_cut:
            continue
        x1.append(drop_dt[ct][k])
        y1.append(mean_dt[ct][k])
        ct_selected.append(ct)
    plt.figure(figsize=(5,5))
    plt.scatter(x1,y1)
    for i, ct in enumerate(ct_selected):
        plt.annotate(ct,(x1[i],y1[i]),fontsize=10)
    plt.title(gene)
    plt.xlim(0,1)
    plt.xlabel("ratio cells detected")
    plt.ylabel("mean exp")
    plt.grid(False)
    plt.show()
    
    return(x1,y1,ct_selected)

def draw_marker_v2(adata, gene, anno_key, cs, show=True,show_cut = 0.2):
    
    from adjustText import adjust_text

    cdm_out = adata.uns['cdm_%s'%(anno_key)]
    
    cnt_dt = cdm_out["cnt"]
    drop_dt = cdm_out["drop"]
    mean_dt = cdm_out["mean"]
    
    ctlist = np.array([x for x in sorted(list(drop_dt.keys())) if x in set(adata.obs[anno_key])])
    
    k = np.where(adata.raw.var_names == gene)[0][0]
    x1 = []
    y1 = []
    ct_selected = []
    cls = []
    for ct in ctlist:
        if cnt_dt[ct][k] < 0:
            continue
        x1.append(drop_dt[ct][k])
        y1.append(mean_dt[ct][k])
        ct_selected.append(ct)
        cls.append(cs[ct])
    plt.figure(figsize=(5,5))
    plt.scatter(x1,y1,color=cls)

    texts = []
    for i, txt in enumerate(ct_selected):
        if ((x1[i] < show_cut) & (y1[i] < show_cut)):
            continue
        texts.append(plt.text(x1[i],y1[i],txt,fontsize=10,color=cs[txt]))
        
    plt.title(gene)
    plt.xlim(0,max(x1)+0.1)
    plt.xlabel("ratio cells detected")
    plt.ylabel("mean exp")
    plt.grid(False)
    adjust_text(texts,only_move={'texts':'xy'},lim=3)
    if show:
        plt.show() 


def draw_marker_blob_v2(adata,cdm_out,genes,ctlist = None,fontsize=10,
                     figscale=0.25,save=None, show=True,
                     min_drop_ratio=0.01,cmap='OrRd',offset=0.5,normed=False):

    def get_drop(ct,drop_dt,min_drop_ratio):
        if ct in drop_dt:
            drop = drop_dt[ct][k]
            if drop > min_drop_ratio:
                return drop
            else:
                return 0
        else:
            return 0
        
    
    cnt_dt = cdm_out["cnt"]
    drop_dt = cdm_out["drop"]
    mean_dt = cdm_out["mean"]
    
    if ctlist == None:
        ctlist = np.array(sorted(list(drop_dt.keys())))

    xpos = []
    ypos = []
    yticks = []
    yticklabels = []

    x1 = []
    y1 = []

    yidx = 0
    
    xpos_list = []
    xticks = []
    offset = 0
    
    for i,ct in enumerate(ctlist):
        if ct not in mean_dt:
            offset -= offset
        else:
            xticks.append(i+offset)
        xpos_list.append(i+offset)
    
    
    for gene in genes[::-1]:

        if gene in adata.raw.var_names:
            yidx += 1
            k = np.where(adata.raw.var_names == gene)[0][0]
            
            ymax = np.max(np.array([mean_dt[ct][k] if ct in mean_dt else 0 for ct in list(mean_dt.keys())]))

            x1.extend([get_drop(ct,drop_dt,min_drop_ratio) for ct in ctlist])
            y0 = np.array([mean_dt[ct][k] if ct in mean_dt else 0 for ct in ctlist])
            if normed:
                y1.extend(list(y0/np.max(y0)))
            else:
                y1.extend(list(y0/ymax))
            yticks.append(yidx)
            yticklabels.append(gene)
            
        else:

            yidx += 0.5
            x1.extend([0]*len(ctlist))
            y1.extend([0]*len(ctlist))
                                   
        xpos.extend(xpos_list)
        ypos.extend([yidx]*len(ctlist))

    x1 = np.array(x1)
    y1 = np.array(y1)
    
    fig = plt.figure(figsize=(len(ctlist)*figscale,len(genes)*figscale*0.9))
    ax = fig.add_subplot(1,1,1)
    sc = ax.scatter(xpos,ypos,s=x1*figscale*400,c=y1,cmap=cmap)
    ax.set_yticks(yticks)
    ax.set_yticklabels(yticklabels,fontsize=fontsize)
    ax.set_xticks(xticks)
    ax.set_xticklabels([x for x in ctlist if x in mean_dt],rotation=90,fontsize=fontsize)
    ax.set_ylim(min(ypos)-0.5,max(ypos)+0.5)
    ax.set_xlim(min(xpos)-0.5,max(xpos)+0.5)
    plt.grid(False)
    
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(sc,cax=cax)
    cbar.ax.tick_params(labelsize=8) 
    
    if save:
        plt.savefig(save,format='pdf')
    if show:
        plt.show()
    
def draw_marker_blob(adata,cdm_out,genes,ctlist = None,fontsize=10,
                     figscale=0.25,save=None, show=True,
                     min_drop_ratio=0.01,cmap='OrRd'):

    cnt_dt = cdm_out["cnt"]
    drop_dt = cdm_out["drop"]
    mean_dt = cdm_out["mean"]
    
    if ctlist == None:
        ctlist = np.array(sorted(list(drop_dt.keys())))

    xpos = []
    ypos = []
    yticks = []
    yticklabels = []

    x1 = []
    y1 = []

    yidx = 0
    
    for gene in genes[::-1]:

        if gene in adata.raw.var_names:
            yidx += 1
            k = np.where(adata.raw.var_names == gene)[0][0]

            x1.extend([drop_dt[ct][k] if ((drop_dt[ct][k] > min_drop_ratio)) else 0 for ct in ctlist])
            y0 = np.array([mean_dt[ct][k] for ct in ctlist])
            y1.extend(list(y0/np.max(y0)))
            
            yticks.append(yidx)
            yticklabels.append(gene)
            
        else:

            yidx += 0.5
            x1.extend([0]*len(ctlist))
            y1.extend([0]*len(ctlist))
                                   
        xpos.extend(list(range(len(ctlist))))
        ypos.extend([yidx]*len(ctlist))

    x1 = np.array(x1)
    y1 = np.array(y1)
    
    fig = plt.figure(figsize=(len(ctlist)*figscale,len(genes)*figscale*0.9))
    ax = fig.add_subplot(1,1,1)
    sc = ax.scatter(xpos,ypos,s=x1*figscale*400,c=y1,cmap=cmap)
    ax.set_yticks(yticks)
    ax.set_yticklabels(yticklabels,fontsize=fontsize)
    ax.set_xticks(range(len(ctlist)))
    ax.set_xticklabels(ctlist,rotation=90,fontsize=fontsize)
    ax.set_ylim(min(ypos)-0.5,max(ypos)+0.5)
    ax.set_xlim(min(xpos)-0.5,max(xpos)+0.5)
    plt.grid(False)
    
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(sc,cax=cax)
    cbar.ax.tick_params(labelsize=8) 
    
    if save:
        plt.savefig(save,format='pdf')
    if show:
        plt.show()
    

def draw_marker_blob_T(adata,cdm_out,genes,ctlist = None,fontsize=10, figscale=0.25,cmap=None,save=None):
    
    if cmap == None:
        cmap='OrRd'

    cnt_dt = cdm_out["cnt"]
    drop_dt = cdm_out["drop"]
    mean_dt = cdm_out["mean"]
    
    if ctlist == None:
        ctlist = np.array(sorted(list(drop_dt.keys())))
    totlist = np.array(sorted(list(drop_dt.keys())))

    xpos = []
    ypos = []
    xticks = []
    xticklabels = []

    x1 = []
    y1 = []

    xidx = 0
    
    ctlist = ctlist[::-1]
    
    for gene in genes:

        if gene in adata.raw.var_names:
            xidx += 1
            k = np.where(adata.raw.var_names == gene)[0][0]

            y1.extend([drop_dt[ct][k] if drop_dt[ct][k] > 0.01 else 0 for ct in ctlist])
            x0 = np.array([mean_dt[ct][k] for ct in ctlist])
            xt = np.array([mean_dt[ct][k] for ct in totlist])
            x1.extend(list(x0/np.max(x0)))
            #x1.extend(list(x0/np.max(xt)))
            
            xticks.append(xidx)
            xticklabels.append(gene)
            
        else:

            xidx += 0.5
            y1.extend([0]*len(ctlist))
            x1.extend([0]*len(ctlist))
                                   
        ypos.extend(list(range(len(ctlist))))
        xpos.extend([xidx]*len(ctlist))

    x1 = np.array(x1)
    y1 = np.array(y1)
    
    
    fig = plt.figure(figsize=(len(genes)*figscale*0.7,len(ctlist)*figscale*0.8))
    ax = fig.add_subplot(1,1,1)
    sc = ax.scatter(xpos,ypos,s=y1*figscale*400,c=x1,cmap=cmap,vmax=1.0)
    ax.xaxis.tick_top()
    ax.set_xticks(np.array(xticks))
    ax.set_xticklabels(xticklabels,rotation=90,fontsize=fontsize)
    ax.set_yticks(range(len(ctlist)))
    ax.set_yticklabels(ctlist,fontsize=fontsize)
    ax.set_ylim(min(ypos)-0.5,max(ypos)+0.5)
    ax.set_xlim(min(xpos)-0.5,max(xpos)+0.5)
    plt.grid(False)
    
    from mpl_toolkits.axes_grid1 import make_axes_locatableg
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="2%", pad=0.05)
    cbar = plt.colorbar(sc,cax=cax)
    cbar.ax.tick_params(labelsize=8) 
    if save != None:
        plt.savefig(save,bbox_inches='tight',format='pdf',dpi=300)
    plt.show()

class volcano_plot():
    '''
    creat volcano plot from anndata
    '''
    
    def __init__(self,adata,anno_key,comp1,comp2,P=0.01,quick=True, normalise=False,
                fc_cut_pval = 0.5, n_pos_pval=5):
        '''
        param P :pseudocount for fc calculation
        '''
        from scipy.stats import ttest_ind
        self.genelist = adata.raw.var_names
        index1 = adata.obs[anno_key]==comp1
        index2 = adata.obs[anno_key]==comp2

        exp1 = adata.raw[index1].X.todense()
        exp2 = adata.raw[index2].X.todense()
        
        self.pval = []
        self.fc = []
        self.n_max = []
        
        if normalise:
            median1 = np.log(np.median(adata.obs['n_counts'][index1]))
            median2 = np.log(np.median(adata.obs['n_counts'][index2]))
        else:
            median1 = 1
            median2 = 1
        
        for i in range(adata.raw.shape[1]):
            
            n_pos1 = np.sum(exp1[:,i].A1>0)
            n_pos2 = np.sum(exp2[:,i].A1>0)
            n_max = np.max([n_pos1,n_pos2])
            
            self.n_max.append(n_max)
            
            count1 = np.mean(exp1[:,i].A1)
            count2 = np.mean(exp2[:,i].A1)
            
            norm_count1 = count1*(1/median1)+P
            norm_count2 = count2*(1/median2)+P
            
            #print(count1, count2, norm_count1, norm_count2)
            
            self.fc.append(np.log2(norm_count1 / norm_count2))
                           
            if quick:
                if np.abs(self.fc[-1])< fc_cut_pval:
                    self.pval.append(1)
                elif n_max < n_pos_pval:
                    self.pval.append(1)
                else:
                    self.pval.append(ttest_ind(exp1[:,i].A1*(1/median1), exp2[:,i].A1*(1/median2))[1])
            else:
                self.pval.append(ttest_ind(exp1[:,i].A1*(1/median1), exp2[:,i].A1*(1/median2))[1])
            
        self.pval = np.array(self.pval)
        self.fc = np.array(self.fc)
        self.n_max = np.array(self.n_max)
            
    def draw(self, title=None, pvalue_cut=100, to_show = 0.2, adjust_lim = 5, show=True, sig_mode = 'auto',adjust = True):
        '''
        draw volcano plot
        param pvalue_cut :-log10Pvalue for cutoff
        sig_mode: ['auto','complex','pval']
        '''
        from adjustText import adjust_text
        plt.figure(figsize=(6,6))

        xpos = np.array(self.fc)
        ypos = -np.log10(np.array(self.pval))
        ypos[ypos==np.inf] = np.max(ypos[ypos!=np.inf])

        if sig_mode == 'complex':
            index = (np.abs(xpos))*ypos
            index_cut = np.percentile(index,100-to_show)
            sig = (np.abs(xpos) > 1) & (ypos > 2) & ((np.abs(xpos))*ypos > index_cut)
        elif sig_mode =='pval':
            sig = (np.abs(xpos) > 1) & (ypos > pvalue_cut)
        elif sig_mode =='auto':
            index_cut = np.percentile(ypos,100-to_show)
            sig = (np.abs(xpos) > 1) & (ypos > index_cut)
        else:
            print('error, check sig_mode')
            raise SystemError

        if title:
            plt.title(title)
        plt.xlabel('log2FoldChange')
        plt.ylabel('-log10Pval')
        plt.scatter(xpos,ypos,s=1, color='k', alpha =0.5, rasterized=True)
        plt.scatter(xpos[sig],ypos[sig],s=3,color='red', rasterized=True)

        texts = []
        for i, gene in enumerate(self.genelist[sig]):
            texts.append(plt.text(xpos[sig][i],ypos[sig][i],gene,fontsize=5))
        
        if adjust:
            adjust_text(texts,only_move={'texts':'xy'},lim=adjust_lim)
        else:
            pass
        if show:
            plt.show() 



import pandas as pd
from adjustText import adjust_text
import numpy as np
import matplotlib.pyplot as plt
from . import *
import sys

c_m = hom['NCBI Taxon ID']==10090
c_h = hom['NCBI Taxon ID']==9606

MtH = {}
HtM = {}
for HomID in sorted(set(hom['HomoloGene ID'])):
    c1 = hom['HomoloGene ID']==HomID
    try: hum = np.array(hom[c1&c_h]['Symbol'])[0]
    except: continue
    try: mus = np.array(hom[c1&c_m]['Symbol'])[0]
    except: continue
    MtH[mus] = hum 
    HtM[hum] = mus
    
def mouse_to_human(x):
    if x in MtH:
        return MtH[x]
    else:
        return x.upper()
    

def draw_scatter_speices_ct(ct,fc,version,fig_folder,percentile_thres=99.5,
                           adjust_lim = 3,fontsize=4):
    v3 = fc[ct]['hs']
    v4 = fc[ct]['mm']

    xpos = np.array(v3.fc) 
    ypos = np.array(v4.fc)

    ymax = np.percentile(ypos,percentile_thres)
    xmax = np.percentile(xpos,percentile_thres)

    sig = (((ypos) > ymax) & (-np.log10(v4.pval) > 5))|\
           (((xpos) > xmax) & (-np.log10(v3.pval) > 5))

    plt.scatter(xpos,ypos,s=1, color='k', alpha =0.5, rasterized=True)
    plt.title(ct)
    texts = []
    for i, gene in enumerate(v3.genelist[sig]):
        texts.append(plt.text(xpos[sig][i],ypos[sig][i],gene,fontsize=fontsize))
    if adjust_lim:
        adjust_text(texts,only_move={'texts':'xy'},lim=adjust_lim)
    plt.grid(False)
    save_fig(version,'FigS1_epi_%s'%ct,fig_folder=fig_folder)
    
    
def draw_scatter_speices_ct_v2(ct,fc,version,fig_folder,percentile_thres=99.5,
                           adjust_lim = 3, fontsize=5):
    v3 = fc[ct]['h_m']
    v4 = fc[ct]['mks']

    xpos = np.array(v3.fc) 
    ypos = np.array(v4.fc)

    ymax = np.percentile(ypos,percentile_thres)
    xmax = np.percentile(np.abs(xpos),percentile_thres)
    
    sig1 = (((ypos) > ymax) & (-np.log10(v4.pval) > 5)) 
    sig2 = ((np.abs(xpos) > xmax) & (-np.log10(v3.pval) > 5))

    sig = sig1
           

    plt.scatter(xpos,ypos,s=1, color='k', alpha =0.5, rasterized=True)
    plt.title(ct)
    texts = []
    for i, gene in enumerate(v3.genelist[sig]):
        texts.append(plt.text(xpos[sig][i],ypos[sig][i],gene,fontsize=fontsize))
    if adjust_lim:
        adjust_text(texts,only_move={'texts':'xy'},lim=adjust_lim)
    plt.grid(False)
    save_fig(version,'FigS1_species_v2_%s'%ct,fig_folder=fig_folder)


# import common modules
from collections import Counter, defaultdict
import os, sys, glob, re
import scipy
import numpy as np
import pandas as pd
import pickle as pkl
import matplotlib.pyplot as plt
import seaborn as sns

# single-cell analysis modules
import scanpy as sc
import scanpy.external as sce
from geosketch import gs
import scrublet as scr
from bbknn import bbknn

# define folders
import site
sitepackagepathway=site.USER_SITE

cwd = sitepackagepathway+'/sceleto-0.8.9-py3.8.egg/sceleto'
data_folder = os.path.join(cwd,'data')
sys.path.append(cwd)

data_folder = os.path.join(cwd,'data')
sys.path.append(cwd)

# internal import

from .genes import cc_genes
from .colors import vega_20, vega_20_scanpy, zeileis_26, godsnot_64
from .markers import find_markers, show_marker_plot
from .model import generate_training_X

from . import network
from . import model
from . import species
from . import markers
from . import utils
from . import genes

# matrix files
#matrix_files = open(os.path.join(cwd,'./matrix/h5ad_files.txt')).read()

# import genesets
tf_df = pd.read_csv(os.path.join(data_folder,'D01_Human_TF.txt'),sep='\t')
tf_genes = list(tf_df.Symbol) # tf gene list (from Animal TF db)
cd_genes = pd.read_csv(os.path.join(data_folder,'D02_CD_genes.txt'),sep='\t') # CD gene list (HUGO)

# General single-cell anlaysis procedure

def sc_process(adata,pid = 'fspkuc',n_pcs=50): # simplified scanpy preprocessing
    '''n: normalise
       l: log
       f: filter hvg
       r: remove cc_genes
       s: scale
       p: pca
       k: knn_neighbors
       u: umap
       c: leiden clusering
       '''
    if 'n' in pid:
        sc.pp.normalize_per_cell(adata,counts_per_cell_after=10e4)
    if 'l' in pid:
        sc.pp.log1p(adata)
        adata.raw = adata
        print('adding raw...')
    if 'f' in pid:
        if adata.raw == None:
            adata.raw = adata
            print('adding raw...')
        sc.pp.filter_genes_dispersion(adata)
    if 'r' in pid:
        adata = remove_geneset(adata,cc_genes)
        print('removing cc_genes...')
    if 's' in pid:
        sc.pp.scale(adata,max_value=10)
    if 'p' in pid:
        sc.pp.pca(adata)
    if 'k' in pid:
        sc.pp.neighbors(adata,n_pcs=n_pcs)
    if 'u' in pid:
        sc.tl.umap(adata)
    if 'c' in pid:
        sc.tl.leiden(adata)
    return adata
    
def read_process(adata,version,
                 species = 'human',
                 sample = None,
                 define_var = True,
                 call_doublet = True,
                 write = True,
                 min_n_counts = 1000,
                 min_n_genes = 500,
                 max_n_genes = 7000,
                 max_p_mito = 0.5
                ):
    if sample:
        adata.obs['Sample'] = sample
    if define_var:
        adata.var['GeneName'] = list(adata.var.gene_ids.index)
        adata.var['EnsemblID'] = list(adata.var.gene_ids)
    adata.obs['n_counts'] = np.sum(adata.X, axis=1).A1
    adata.obs['n_genes'] = np.sum(adata.X>0,axis=1)
    
    print('calculating mito... as species = {}'.format(species))
    if species=='mouse':
        mito_genes = adata.var_names.str.startswith('mt-')
        adata.obs['mito'] = (np.sum(adata.X[:, mito_genes],axis=1).A1) / ((np.sum(adata.X,axis=1).A1)+1)
    elif species=='human':
        mito_genes = adata.var_names.str.startswith('MT-')
        adata.obs['mito'] = (np.sum(adata.X[:, mito_genes],axis=1).A1) / ((np.sum(adata.X,axis=1).A1)+1)
    else:
        print("check_species: weird mito gene names")
        raise SystemError
        
    print('filtering cells... higher than {} counts, more than {} and less than {} genes, less than {} p_mito...'.format(min_n_counts, min_n_genes, max_n_genes, max_p_mito))
    # filter cells
    clist = []
    clist.append(np.array(adata.obs['n_counts'] > min_n_counts))
    clist.append(np.array(adata.obs['n_genes'] > min_n_genes))
    clist.append(np.array(adata.obs['n_genes'] < max_n_genes))
    clist.append(np.array(adata.obs['mito'] < max_p_mito))
    
    c = np.column_stack(clist).all(axis=1)
    adata = adata[c].copy()
    
    if call_doublet:
        print('calling doublets using scrublet...')
        scrub = scr.Scrublet(adata.X)
        doublet_scores, predicted_doublets = scrub.scrub_doublets(verbose=False)
        adata.obs['doublet_scores'] = doublet_scores
        adata.obs['predicted_doublets'] = predicted_doublets
    
    if write:
        print('writing output into write/%s%s_filtered.h5ad ...'%(version,sample))
        sc.write('%s%s_filtered'%(version,sample),adata)
    return adata

def write_notebook(name1,name2):
    os.system(f'jupyter nbconvert {name1} --to notebook --ClearOutputPreprocessor.enabled=True --output {name2}')
        
def sort_var_names_based_on_GeneID(adata):
    return adata[:,np.argsort(adata.var.GeneID)].copy()

def combine_batch(adata,key1,key2,new_key = 'batch'):
    print('storing new batch into '+new_key)
    adata.obs[new_key] = ['{}_{}'.format(k1,k2) for k1,k2 in zip(adata.obs[key1],adata.obs[key2])]

def doublet(adata, key='Sample'):
    '''detecting doublet using scrublet per key'''
    doublet = []
    for filename in set(adata.obs[key]):
        print(filename)
        sdata = adata[adata.obs[key] == filename].copy()
        scrub = scr.Scrublet(sdata.X)
        doublet_scores, predicted_doublets = scrub.scrub_doublets(verbose=False)
        doublet.extend([(x,y,z) for x,y,z in zip(sdata.obs_names,doublet_scores,predicted_doublets)])
    doublet_score = {x:y for (x,y,z) in doublet}
    doublet_predict = {x:z for (x,y,z) in doublet}
    adata.obs['doublet_score'] = [doublet_score[obs_name] for obs_name in list(adata.obs_names)]
    adata.obs['doublet_predict'] = [doublet_predict[obs_name] for obs_name in list(adata.obs_names)]

def get_sketch(adata,key,folds=10,how='pd',min_num_per_key=500,start='filter',raw=True):
    '''geometric sketching based on diffusion map and pca
    folds: folds to subsample
    min_num_per_key: minimun number to sample'''
    sketch_index = []
    for smp in set(adata.obs[key]):
        print(smp)
        c = adata.obs[key] == smp
        print('from:',sum(c))
        
        if start=='filter':
            sdata = get_subset(adata,c,raw=raw)
        else:        
            sdata = adata[c]
            sc.pp.filter_genes_dispersion(sdata)
            sc.pp.pca(sdata)
        
        if 'd' in how:
            sc.pp.neighbors(sdata)
            sc.tl.diffmap(sdata)

        N = np.max([np.int(np.sum(c)/folds),np.min([min_num_per_key,np.sum(c)])])
        print('to select:',N)
        if how =='pd':
            set1 = set(sdata.obs_names[gs(sdata.obsm['X_diffmap'],N,replace=False)])
            set2 = set(sdata.obs_names[gs(sdata.obsm['X_pca'][:,:50],N,replace=False)])
            sketch_index.extend(list(set1.union(set2)))
        elif how =='p':
            set2 = set(sdata.obs_names[gs(sdata.obsm['X_pca'][:,:50],N,replace=False)])
            sketch_index.extend(list(set2))
        elif how =='d':
            set1 = set(sdata.obs_names[gs(sdata.obsm['X_diffmap'][:,:20],N,replace=False)])
            sketch_index.extend(list(set1))
        else:
            raise SystemError
        print('length of sketch:',len(sketch_index))
    return(sketch_index)
        
def bbknn_umap(adata,batch_key,n_pcs,cluster=False,n_neighbors=3):
    bbknn(adata,batch_key=batch_key,n_pcs=n_pcs,approx=False,neighbors_within_batch=n_neighbors)
    if cluster:
        sc.tl.leiden(adata)
    sc.tl.umap(adata)
    
def umap(adata,name=None):
    sc.tl.umap(adata)
    if name:
        adata.obsm['X_umap_'+name] = adata.obsm['X_umap'].copy()
        
def umap_show(adata,feature,feature_name= None):
    if feature_name:
        adata.obs[feature_name] = feature
        sc.pl.umap(adata,color=feature_name,color_map='OrRd')
    else:
        adata.obs['show'] = feature
        sc.pl.umap(adata,color='show',color_map='OrRd')

# Clustering
        
def leiden_res(adata,res,show=False):
    print('calculating leiden at res {0:.2f}...'.format(res))
    sc.tl.leiden(adata,resolution=res)
    print('copying into obs.leiden_{0:.2f}...'.format(res))
    adata.obs['leiden_{0:.2f}'.format(res)] = adata.obs['leiden'].copy()
    if show:
        sc.pl.umap(adata,color='leiden_{0:.2f}'.format(res))
        
def subcluster(adata,obs_label,cl_label,new_label,res=0.1):
    '''
    take specific cluster from adata and split that into smaller cluster
    adata: AnnData object
    obs_label: obs label. eg. 'leiden' or 'celltype'
    cl_label: cluster name. eg. '1' or 'macrophage'
    new_label: name to store updated label
    '''
    subset = adata[adata.obs[obs_label]==cl_label].copy()
    sc.tl.leiden(subset,resolution=res)
    update_dict = {obs_name:cl_label+'_'+new_cl for obs_name,new_cl in zip(subset.obs_names,subset.obs['leiden'])}
    adata.obs[new_label] = [old if obs_name not in update_dict else update_dict[obs_name] for obs_name, old in zip(adata.obs_names,adata.obs[obs_label])]
        
def remove_geneset(adata,geneset):
    adata = adata[:,~adata.var_names.isin(list(geneset))].copy()
    return adata

def is_cycling(adata,cc_genes=cc_genes,cut_off=0.4):
    X = np.mean(adata.raw.X[:,adata.raw.var_names.isin(cc_genes)],axis=1)
    plt.hist(X)
    adata.obs['Cycle_score'] = X
    adata.obs['isCycle'] = X>cut_off
    
def get_subset(idata, select, cc_genes=cc_genes, log=False,raw=True):
    if raw:
        adata = sc.AnnData(idata[select].raw.X)
        adata.var = idata.raw.var
    else:
        adata = sc.AnnData(idata[select].X)
        adata.var = idata.var
    adata.obs = idata.obs[select]
    adata.raw = adata.copy()
    #adata.X = scipy.sparse.csr_matrix(np.exp(adata.X.todense())-1)
    sc.pp.filter_genes_dispersion(adata,log=log)
    if log:
        sc.pp.log1p(adata)
    sc.pp.scale(adata,max_value=10)
    if len(cc_genes)>0:
        print('removing cc_genes...')
        adata = remove_geneset(adata,cc_genes)
    sc.pp.pca(adata,n_comps = np.min([50,adata.X.shape[0],adata.X.shape[1]]))
    return adata

def get_raw(idata):
    adata = sc.AnnData(idata.raw.X)
    adata.var = idata.raw.var
    adata.obs = idata.obs
    adata.raw = adata.copy()

    return adata

def get_raw_process(idata, cc_genes=cc_genes, log=False):
    adata = sc.AnnData(idata.raw.X)
    adata.var = idata.raw.var
    adata.obs = idata.obs
    adata.raw = adata.copy()
    #adata.X = scipy.sparse.csr_matrix(np.exp(adata.X.todense())-1)
    sc.pp.filter_genes_dispersion(adata,log=log)
    if log:
        sc.pp.log1p(adata)
    sc.pp.scale(adata,max_value=10)
    if len(cc_genes)>0:
        print('removing cc_genes...')
        adata = remove_geneset(adata,cc_genes)
    sc.pp.pca(adata,n_comps = np.min([50,adata.X.shape[0],adata.X.shape[1]]))
    return adata

def output_matrix_Seurat(adata,version,name,use_raw=False):
    from scipy.io import mmwrite
    
    if use_raw:
        X = adata.raw.X
        mmwrite(version+name+'.mtx',X)
        adata.obs.to_csv(version+name+'.meta.csv')
        adata.raw.var.to_csv(version+name+'.var.csv')
    else:
        X = adata.X
        mmwrite(version+name+'.mtx',X)
        adata.obs.to_csv(version+name+'.meta.csv')
        adata.var.to_csv(version+name+'.var.csv')

def us(adata,gene,groups=None, show=False, exclude =None,figsize=None,**kwargs):
    from matplotlib import rcParams
    if figsize:
        rcParams['figure.figsize'] = figsize
    if ',' in gene:
        gene = gene.split(',')
    if groups:
        sc.pl.umap(adata,color=gene,color_map='OrRd',groups=groups, show=show, **kwargs)
    else:
        if exclude: # list to not show
            groups = [x for x in set(adata.obs[gene]) if x not in exclude]
            sc.pl.umap(adata,color=gene,color_map='OrRd',groups=groups, show=show, **kwargs)
        else:
            sc.pl.umap(adata,color=gene,color_map='OrRd',show=show, **kwargs)
    rcParams['figure.figsize'] = [5,5]
    
def merge_matrix(ad,obskeys = None,use_raw = False,keep_only_mutual=False):
    '''merge matrix stored in ad
    ad: dictionary of anndata to merge
    obskeys: list to merge within anndata
    use_raw: if True, merge from .raw.X'''
    
    smp_list = list(ad.keys())
    obs_dict = defaultdict(list)
    obs_names = []
    
    for smp in smp_list:
        ad[smp].obs['name'] = smp
    
    if not obskeys:
        obskey_list = []
        obskeys = []
        for sample in smp_list:
            obskey_list.extend(list(ad[sample].obs.columns))
        for (obskey, number) in Counter(obskey_list).items():
            if number == len(smp_list):
                obskeys.append(obskey)
            else:
                if keep_only_mutual:
                    pass
                else:
                    for sample in smp_list:
                        if obskey not in ad[sample].obs.columns:
                            ad[sample].obs[obskey]='n/a'
                    obskeys.append(obskey)
                               
    for sample in smp_list:
        obs_names.extend(list(ad[sample].obs_names))
        for key in obskeys:   
            obs_dict[key].extend(list(ad[sample].obs[key]))
    
    from scipy.sparse import vstack
    if use_raw == True:
        stack = vstack([ad[x].raw.X for x in smp_list]) # stack data
        adata = sc.AnnData(stack, var = ad[smp_list[0]].raw.var)
    else:
        stack = vstack([ad[x].X for x in smp_list]) # stack data
        adata = sc.AnnData(stack, var = ad[smp_list[0]].var)
      
    
    adata.obs_names = obs_names
    print(len(adata))
    for obs_col in obs_dict:
        print(obs_col)
        adata.obs[obs_col] = obs_dict[obs_col]
    return adata

def timestamp():
    from datetime import datetime
    return datetime.now().strftime("%y%m%d%H%M")

def save_html(name,log=False): # export notebook into html file
    time = timestamp()
    print(os.system('jupyter nbconvert --to html %s'%(name)))
    name_key = re.sub('.ipynb$','',name)
    if log:
        print(os.system('mv %s.html %s_%s_%s.html'%(name_key,name_key,time,log)))
    else:
        print(os.system('mv %s.html %s_%s.html'%(name_key,name_key,time)))

def write(adata,version,name):
    '''write adata into [name]'''
    name = version + name
    sc.write(name,adata)
    print("_".join(name.split(".")) + " = '%s'"%name)
    
def save_fig(version,figcount,fig_format='pdf',fig_folder='11_Figs'):
    
    plt.savefig('%s/%s%s.%s'%(fig_folder,version,figcount,fig_format),bbox_inches='tight',format=fig_format,dpi=300)
    print('%s/%s%s.pdf'%(fig_folder,version,figcount))

# batch regression methods

def regress_batch_v2(adata,batch_key,confounder_key):
    '''batch regression tool
    batch_key=list of observation categories to be regressed out
    confounder_key=list of observation categories to be kept
    returns ndata with corrected X'''

    from sklearn.linear_model import Ridge
    
    print('fitting linear model...')
    dummy = pd.get_dummies(adata.obs[batch_key+confounder_key],drop_first=False)
    X_exp = adata.X # scaled data
    if scipy.sparse.issparse(X_exp):
        X_exp = X_exp.todense()
    LR = Ridge(fit_intercept=False,alpha=1.0)
    LR.fit(dummy,X_exp)

    if len(batch_key)>1:
        batch_index = np.logical_or.reduce(np.vstack([dummy.columns.str.startswith(x) for x in batch_key]))
    else:
        batch_index = np.vstack([dummy.columns.str.startswith(x) for x in batch_key])[0]
    
    print('corrcting batch...')
    dm = np.array(dummy)[:,batch_index]
    X_explained = dm.dot(LR.coef_[:,batch_index].T)
    X_remain = X_exp - X_explained
    ndata = sc.AnnData(X_remain)
    ndata.obs = adata.obs
    ndata.var = adata.var
    return ndata, X_explained

def regress_iter(adata,batch_key,confounder_key,bbknn_key,scale=True, approx = True,n_pcs=50):
    if scale == True:
        print('scaling data...')
        sc.pp.scale(adata,max_value=10)
    ndata, X_explained = regress_batch_v2(adata,batch_key=batch_key,confounder_key=confounder_key)
    print('running pca...')
    sc.pp.pca(ndata)
    print('running bbknn...')
    bbknn(ndata, batch_key = bbknn_key,n_pcs=n_pcs, approx=approx)
    return ndata #, X_explained

def run_pca_bbknn_umap(ad,level_key,bbknn_key,marker_dict,
                       resolution=0.02,start = 'leiden',select=False,
                      thres=0.95,min_drop_cut=0.5,show=True, how='almost',
                      min_cluster_num = 200):
    '''
    run pca, bbknn, umap and clustering to find good low-rescluster with markers
    how = [almost, any, all] for marker_found function
    '''
    adata = ad[level_key]
    if start == 'pca':
        sc.pp.pca(adata)
        bbknn(adata,batch_key=bbknn_key,approx=False)
        sc.tl.umap(adata)
    if start in 'leiden,pca'.split(','):
        sc.tl.leiden(adata,resolution=resolution)
        adata.obs[level_key] = list(adata.obs['leiden'])
    if start in 'leiden,pca,mks'.split(','):
        if show:
            sc.pl.umap(adata,color='leiden')
        if len(set(adata.obs['leiden']))<2:
            print('clustering not enough')
            return False,None
        elif np.min([x for x in Counter(adata.obs['leiden']).values()]) < min_cluster_num:
            print('clustering resolution too high')
            return False,None
        else:
            if select:
                ndata = generate_training_X(adata,'leiden',select_num=select)
            else:
                ndata = adata
            mks = find_markers(ndata,'leiden',thres=thres,min_drop_cut=min_drop_cut)
            if show:
                show_marker_plot(ndata,'leiden',mks,toshow=5)
            go = marker_found(mks,how=how)
            if go:
                commit_level(adata,level_key,mks,marker_dict)
            else:
                print('marker not found')
            return go, mks
    else:
        'start accepts either pca or leiden'
        raise SystemError

def marker_found(mks,how='any'):
    if how=='any':
        c1 = len([keys for keys,values in mks.items() if len(values)>0]) > 0 
    elif how == 'some':
        c0 = len([keys for keys,values in mks.items() if len(values)>0]) >= 3
        c2 = len([keys for keys,values in mks.items() if len(values)>0]) >= (len(mks.keys())-1)
        c1 = c0|c2
    elif how=='all':
        c1 = len([keys for keys,values in mks.items() if len(values)>0]) == len(mks.keys())
    elif how=='almost':
        c1 = len([keys for keys,values in mks.items() if len(values)>0]) >= (len(mks.keys())-1)
    else:
        print('Error: print how not in any, all, alomst')
        raise SystemExit
    return c1

def commit_level(adata,level_key,mks,marker_dict):
    for leiden_clst in set(adata.obs[level_key]):
        to_merge = np.array(adata.obs[level_key].copy(),dtype=object)
        if len(mks[leiden_clst]) >0:
            final_key = level_key+"_"+leiden_clst
            marker_dict[final_key] = mks[leiden_clst]
        else:
            to_merge[adata.obs[level_key]==leiden_clst]='M'
        adata.obs[level_key] = to_merge
                
def expand_level_copy(ad,level_key):
    adata = ad[level_key]
    for leiden_clst in set(adata.obs[level_key]):
        final_key = level_key+"_"+leiden_clst
        print(final_key)
        ad[final_key] = adata[adata.obs[level_key]==leiden_clst].copy()
        
def summary(ad):
    anno = np.zeros(len(ad['0']),dtype=object)
    final_clusters = []
    for k in ad.keys():
        if np.sum([x.startswith(k) for x in ad.keys()]) ==1:
            final_clusters.append(k)
    for k in final_clusters:
        anno[ad['0'].obs_names.isin(ad[k].obs_names)] =k
    return anno
        
def walk_cluster(ad,marker_dict,tried,bbknn_key,
                 leiden_walk=[0.02,0.05], thres=0.95, min_drop_cut=0.5,
                 select=False, show=False, how='almost', 
                 final_limit_num=8, min_num_split=500):
    go = False
    processed = set(['_'.join(x.split('_')[:-1]) for x in marker_dict.keys()])
    to_process = [level_key for level_key in list(ad.keys()) if level_key not in processed.union(set(tried))]

    print(to_process)
    for level_key in to_process:
        print(level_key)
        if len(level_key.split("_")) > final_limit_num:
            print('level too deep')
            continue
        if len(ad[level_key])<min_num_split:
            print('subset too small')
            continue
        for resolution in leiden_walk:
            print(resolution)
            result = run_pca_bbknn_umap(ad,level_key,bbknn_key,marker_dict,
                                             start='leiden',resolution=resolution,
                                            thres=thres,min_drop_cut=min_drop_cut,
                                            select=select,show=show,how=how)
            if result[0]:
                print('marker found at '+str(resolution))
                go = True
                expand_level_copy(ad,level_key)
                break
        tried.append(level_key)
    return(go)

def run_harmony(sdata,vars_use,n_pcs=30):
    meta_data = sdata.obs
    data_mat = sdata.obsm['X_pca']

    import harmonypy as hm
    ho = hm.run_harmony(data_mat,meta_data,vars_use)
    
    sdata.obsm['X_pca_before'] = sdata.obsm['X_pca']
    sdata.obsm['X_pca'] = ho.Z_corr.T
    
    print('calculating neighbors...')
    sc.pp.neighbors(sdata,n_pcs=n_pcs)
    print('calculating umap...')
    sc.tl.umap(sdata)

# easy annotation class
        
class annotater():
    '''
    create de novo annotation onto adata
    '''
    
    def __init__(self,adata,new_label_name,old_label=None):
    
        if old_label:
            adata.obs[new_label_name] = adata.obs[old_label]
        else:
            adata.obs[new_label_name] = 'unknown'
        arr = np.array(adata.obs[new_label_name],dtype=object)
        self.new_label = arr
        self.new_label_name = new_label_name
        
    def update(self,adata,obskey,select,label_name,unknown=False):
        if type(select)==str:
            if ',' in select:
                label_condition = adata.obs[obskey].isin(select.split(','))
            else:
                label_condition = adata.obs[obskey]==select
        else:
            label_condition = adata.obs[obskey]==select
        if unknown:
            print('updating only unknown values...')
            label_condition = label_condition & (adata.obs[self.new_label_name]=='unknown')
        self.new_label[label_condition] = label_name
        adata.obs[self.new_label_name] = self.new_label
    
    def update_condi(self,adata,label_condition,label_name):

        self.new_label[label_condition] = label_name
        adata.obs[self.new_label_name] = self.new_label

def read_files(filename, sample):
    '''import 10X data, based on filename (path to file) and sample ID (assigned as unique ID)'''

    path = '%s/'%(filename)
    adata = sc.read(path+'matrix.mtx',cache=True).transpose()
  
    try: 
        adata.var_names = np.genfromtxt(path + 'genes.tsv',dtype=str)[:,1]
        adata.var['GeneName'] = np.genfromtxt(path + 'genes.tsv', dtype=str)[:, 1]
        adata.var['GeneID'] = np.genfromtxt(path + 'genes.tsv', dtype=str)[:, 0]
    except: 
        adata.var_names = np.genfromtxt(path + 'features.tsv.gz',dtype=str)[:,1]
        adata.var['GeneName'] = np.genfromtxt(path + 'features.tsv.gz', dtype=str)[:, 1]
        adata.var['GeneID'] = np.genfromtxt(path + 'features.tsv.gz', dtype=str)[:, 0]
    adata.obs_names = np.genfromtxt(path + 'barcodes.tsv',dtype=str)
    adata.obs_names = [filename+"-"+x.strip("-1") for x in adata.obs_names]
    adata.obs['Sample'] = sample
    
    # caculate n_counts / n_genes per cell
    adata.obs['n_counts'] = np.sum(adata.X, axis=1).A1
    adata.obs['n_genes'] = np.sum(adata.X>0,axis=1)
    
    mito_genes = adata.var_names.str.startswith('MT-')
    adata.obs['mito'] = (np.sum(adata.X[:, mito_genes],axis=1).A1) / (np.sum(adata.X,axis=1).A1)
    
    # filter cells
    clist = []
    clist.append(np.array(adata.obs['n_counts'] > 1000))
    clist.append(np.array(adata.obs['n_genes'] > 500))
    clist.append(np.array(adata.obs['n_genes'] < 7000))
    clist.append(np.array(adata.obs['mito'] < 0.5))

    c = np.column_stack(clist).all(axis=1)
    adata = adata[c].copy()

    sc.write('%s%s_filtered'%(version,sample),adata)
    
    return adata

def read_files_multi(file_lists, n_pool = 10):
    '''file_lists: list of tuple(filename, sampleID)'''
    
    import multiprocessing as mp

    with mp.Manager() as manager:
        with manager.Pool(n_pool) as pool:
            pool.starmap(read_files, file_lists)

# Remove doublets

def final_doublets(adata,doublet_key = 'doublet_final',fracDoublet = 0.1,leiden_key = 'leiden'):
    doublet_annotator = annotater(adata, doublet_key)

    for cl in set(adata.obs[leiden_key]):
        isCluster = adata.obs[leiden_key]==cl
        nTotal = np.sum(isCluster)
        try: nDoublet = np.sum(adata.obs['predicted_doublets'][isCluster])
        except: nDoublet = np.sum(adata.obs['predicted_doublets'][isCluster]=='True')
        print(nDoublet,nTotal)
        if nDoublet > fracDoublet*nTotal:
            doublet_annotator.update(adata,leiden_key,cl,'doublet')

# Cell composition analysis

def get_crosstab(adata,tab1,tab2):
    df = pd.crosstab(adata.obs[tab1],adata.obs[tab2])
    df_norm1 = df.div(df.sum(axis=0), axis=1)*100
    df_norm2 = df_norm1.div(df_norm1.sum(axis=1), axis=0)*100
    return df_norm1,df_norm2
            
# linear regression for diff_exp
        
class linear_regression():
    
    def __init__(self,adata):
        self.adata = adata
        self.df = adata.obs
        self.X_exp = adata.X
        self.LR_dict = {}
        
    def ridge(self,keys):
        from sklearn.linear_model import Ridge
        cat = self.df[keys]
        LR = Ridge()
        dummy = pd.get_dummies(cat,drop_first=False)
        LR.fit(dummy, self.X_exp)
        print('fitting linear model...')
        params = list(dummy.columns)
        ct_dict = {'LR': LR, 'params': params}
        self.LR_dict['ridge'] = ct_dict
        
    def lasso(self,keys):
        from sklearn.linear_model import Lasso
        cat = self.df[keys]
        LR = Lasso()
        dummy = pd.get_dummies(cat,drop_first=False)
        LR.fit(dummy, self.X_exp)
        print('fitting linear model...')
        params = list(dummy.columns)
        ct_dict = {'LR': LR, 'params': params}
        self.LR_dict['lasso'] = ct_dict
        
    def celltype_key(self, anno_key, celltype, others=['organ','method']):
        from sklearn.linear_model import Ridge
        
        self.df['is_%s'%celltype] = [str(x) for x in self.df[anno_key]==celltype] # cell type term
        cat = self.df[['is_%s'%celltype] + others] # add organ term and others
        LR = Ridge()
        print('fitting linear model...')
        dummy = pd.get_dummies(cat, drop_first=False)
        LR.fit(dummy, self.X_exp)
        params = list(dummy.columns)
        ct_dict = {'LR': LR, 'params': params}
        self.LR_dict[celltype] = ct_dict
  
        
    def celltype_organ(self, anno_key, celltype, others=['organ','method']):
        from sklearn.linear_model import Ridge
        
        self.df['is_%s'%celltype] = [str(x) for x in self.df[anno_key]==celltype] # cell type term
        self.df['%s_organ'%celltype] = [x+'_'+y for x,y in zip(self.df['is_%s'%celltype],self.df['organ'])] # interaction term
        cat = self.df[['is_%s'%celltype, '%s_organ'%celltype] + others] # add organ term and others
        LR = Ridge()
        dummy = pd.get_dummies(cat, drop_first=False)
        LR.fit(dummy, self.X_exp)
        params = list(dummy.columns)
        ct_dict = {'LR': LR, 'params': params}
        self.LR_dict[celltype] = ct_dict
        
    def celltype_organ_2(self, anno_key, celltype, others=['method']):
        from sklearn.linear_model import Ridge
        
        self.df['is_%s'%celltype] = [str(x) for x in self.df[anno_key]==celltype] # cell type term
        self.df['%s_organ'%celltype] = [x+'_'+y for x,y in zip(self.df['is_%s'%celltype],self.df['organ'])] # interaction term
        cat = self.df[['%s_organ'%celltype] + others] # add organ term and others
        LR = Ridge()
        dummy = pd.get_dummies(cat, drop_first=False)
        LR.fit(dummy, self.X_exp)
        params = list(dummy.columns)
        ct_dict = {'LR': LR, 'params': params}
        self.LR_dict[celltype] = ct_dict
    
    def show_param_genes(self, celltype, param, toshow=20, output=False):
        coef = self.LR_dict[celltype]['LR'].coef_[:,self.LR_dict[celltype]['params'].index(param)]
        order = np.argsort(-coef)
        names = self.adata.var_names[order][:toshow]
        values = coef[order][:toshow]
        print(['%s:%.2f'%(x,y) for x,y in zip(names,values)])
        if output:
            return names, values
        
    def get_param_values(self, celltype, param):
        coef = self.LR_dict[celltype]['LR'].coef_[:,self.LR_dict[celltype]['params'].index(param)]
        return coef
    
    def param_summary(self, celltype, gene, show=True,**kwargs):
        LR = self.LR_dict[celltype]['LR']
        params = self.LR_dict[celltype]['params']

        gidx = self.adata.var_names==gene

        cf = LR.coef_[gidx][0]
        cf_idx = np.argsort(-cf)
        print('\n'.join(['%.2f : %s'%(a,b) for a,b in zip(cf[cf_idx],np.array(params)[cf_idx])]))

    def violin_plot(self, celltype, gene, key='organ',show=True,**kwargs):
        LR = self.LR_dict[celltype]['LR']
        params = self.LR_dict[celltype]['params']
        
        Exp = self.adata.raw.X
        exp = Exp[:,self.adata.raw.var_names==gene].todense().A1

        self.df[gene+'_exp'] = exp

        fig = plt.figure(figsize=(8,2))
        ax = plt.subplot(111)
        sns.violinplot(x='is_%s'%(celltype),y=gene+'_exp',
                       hue=key,data = self.df,**kwargs,
                       scale='width',linewidth=0,inner=None,rasterized=True,cut=0,ax=ax)
        plt.xticks(rotation=0)
        plt.grid(False)
        ax.legend(bbox_to_anchor=(1.2, 1.05))
        if show:
            plt.show()

        gidx = self.adata.var_names==gene

        cf = LR.coef_[gidx][0]
        cf_idx = np.argsort(-cf)
        print('\n'.join(['%.2f : %s'%(a,b) for a,b in zip(cf[cf_idx],np.array(params)[cf_idx])]))



from setuptools import setup, find_packages

setup(
name='sceleto',
license='LICENSE',
#url='NULL'
author='Jongeun Park',
author_email='jp24@kaist.ac.kr',
description='Tool to aid in single cell analysis (temporary description)',
#####Package#####
include_package_data=True,
python_requires='>=3.7',
packages=find_packages(include=['sceleto','sceleto.*','.']),
#####Needed dependencies#####
install_requires=[
   'pandas',
   'numpy',
   'scanpy>=1.6.1',
   'scipy',
   'seaborn',
   'networkx',
   'python-igraph==0.9.1',
   'bbknn==1.4.1',
   'geosketch',
   'scrublet',
   'joblib',
   'datetime',
   'harmonypy',
   'matplotlib',
   'geosketch',
   'scrublet',
   'adjustText',
   'numba==0.52.0'
],
#####Sharing#####
version='0.9.3',
#####Search#####
keywords=['sceleto', 'single cell', 'scRNA-seq'],
#####Read Me#####
long_description=open('README.md').read(),
long_description_content_type='text/markdown',
zip_safe=False,
)



import numpy as np
from collections import Counter
from collections import defaultdict
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
import scanpy as sc

# Logistic model methods
    
def generate_training_X(adata,ct_key,select_num=200,exclude=[]):
    ad_model = adata
    ct_list = [x for x in set(ad_model.obs[ct_key])]
    select = np.hstack([np.random.choice(np.where(ad_model.obs[ct_key]==ct)[0],select_num) for ct in ct_list if ct not in exclude])
    ad_model = ad_model[select]
                                         
    return ad_model

def logistic_model(data, cell_types, sparsity=0.2, fraction=0.2, penalty='l2'):
    X = data
    X_train, X_test, y_train, y_test = \
    train_test_split(X, cell_types, test_size=fraction)
    lr = LogisticRegression(penalty=penalty, C=sparsity)
    lr.fit(X_train, y_train)
    y_prob = lr.predict_proba(X_test)
    return y_prob, y_test, lr

from sklearn import metrics
from scanpy.plotting.palettes import *

def plot_roc(y_prob, y_test, lr):
    aucs =[]
    if len(lr.classes_)<21: colors = default_20
    elif len(lr.classes_)<27: colors = default_28
    else: colors = default_102
    for i, cell_type in enumerate(lr.classes_):
        fpr, tpr, _ = metrics.roc_curve(y_test == cell_type, y_prob[:, i])
        auc = metrics.auc(fpr, tpr)
        aucs.append(auc)
        plt.plot(fpr, tpr, c=colors[i], lw=2, label = cell_type)
    plt.plot([0, 1], [0, 1], color='k', ls=':')
    plt.legend(loc=(1,0))
    min_auc = np.min(aucs)
    plt.title("Min AUC: %.3f"%(min_auc))
    return(min_auc)

def transfer_annotation_jp(input_adata,y_id,output_adata,y_out,select_num=200,log=None,exclude=[],raw=True):
    ad_model = generate_training_X(input_adata,y_id,select_num=select_num,exclude=exclude)
    #sc.pl.umap(ad_model,color=y_id)

    if raw==False:
        X_model = ad_model.X
    else:
        check_raw_exists(ad_model)
        X_model = ad_model.raw.X
    y_model = ad_model.obs[y_id]
    if raw==False:
        X_predict = output_adata.X
    else:
        check_raw_exists(output_adata)
        X_predict = output_adata.raw.X

    print("creating lr model...")
    y_prob, y_test, lr =  logistic_model(X_model, y_model, sparsity=0.2, fraction=0.2, penalty='l2')
    plot_roc(y_prob, y_test, lr)
    
    print("making lr prediction...")
    Lout = {}
    if log:
        Lout['log'] = log
    Lout['classes'] = lr.classes_
    Lout['predict'] = lr.predict(X_predict)
    Lout['predict_proba'] = lr.predict_proba(X_predict)
    
    print("updating lr to adata...")
    output_adata.obs[y_out] = Lout['predict']
    
    return lr

def update_label(from_adata,from_label,to_adata,old_label,new_label,
                 exclude=None,include=None,replace=False,unknown=None,keep_replaced=True):
    if old_label not in to_adata.obs.columns:
        to_adata.obs[old_label] = 'unknown'
    if exclude:
        ON = {O:N for O,N in zip(from_adata.obs_names,from_adata.obs[from_label]) if N not in exclude}
    elif include:
        ON = {O:N for O,N in zip(from_adata.obs_names,from_adata.obs[from_label]) if N in include}
    else:
        ON = {O:N for O,N in zip(from_adata.obs_names,from_adata.obs[from_label])}
        
 
    if unknown:
        new_anno = [ON[O] if ((O in ON) & (N == unknown)) else N for O,N in zip(to_adata.obs_names,to_adata.obs[old_label])]
    else:
        new_anno = [ON[O] if (O in ON) else N for O,N in zip(to_adata.obs_names,to_adata.obs[old_label])]
    
    if new_label in to_adata.obs.columns:
        if replace==False:
            if keep_replaced==True:
                raise SystemError
            else:
                to_adata.obs[new_label] = new_anno
        elif replace==True:
            i =  1
            while True:
                new_key = new_label+'.replaced.'+str(i)
                if new_key in to_adata.obs.columns:
                    i+=1
                    continue
                else:
                    to_adata.obs[new_key] = list(to_adata.obs[new_label])
                    del to_adata.obs[new_label]
                    to_adata.obs[new_label] = new_anno
                    break
    else:
        to_adata.obs[new_label] = new_anno

def get_common_var_raw(a,b):
    common = sorted(list(set(a.raw.var_names).intersection(set(b.raw.var_names))))
    list_a_names = list(a.raw.var_names)
    list_b_names = list(b.raw.var_names)
    a_index = np.array([list_a_names.index(x) for x in common])
    b_index = np.array([list_b_names.index(x) for x in common])
    print('calculating a...')
    a_new_X = a.raw.X[:,a_index]
    print('calculating b...')
    b_new_X = b.raw.X[:,b_index]
    a_new = sc.AnnData(a_new_X,obs = a.obs)
    a_new.obsm = a.obsm
    a_new.var_names = common
    b_new = sc.AnnData(b_new_X,obs = b.obs)
    b_new.obsm = b.obsm
    b_new.var_names = common
    return a_new,b_new

def check_raw_exists(adata):
    try: 
        adata.raw.X
        print('raw exist for adata')
    except:
        print("raw not found from adata, adding raw...".format(adata))
        adata.raw = adata

def predict_high(lr, adata, out_name, cl_to_focus = None, p = 0.9):
    out_name = '{}_{}'.format(out_name,str(p))
    pb = lr.predict_proba(adata.raw.X)
    adata.obs[out_name] = 'None'
    if cl_to_focus:
        cl_index = lr.classes_==cl_to_focus
        adata.obs[out_name][(pb[:,cl_index]>p).T[0]] = cl_to_focus
    else:
        for i,cl in enumerate(lr.classes_):
            adata.obs[out_name][pb[:,i]>p] = cl
        
def fill_columns(a_new,b_new):
    a_col_list = list(a_new.obs.columns)
    b_col_list = list(b_new.obs.columns)
    for obs_item in a_col_list:
        if obs_item not in b_col_list:
            b_new.obs[obs_item] = 'nan'
    for obs_item in b_col_list:
        if obs_item not in a_col_list:
            a_new.obs[obs_item] = 'nan'
            
def remove_minor_anno(adata,obskey,num_cut=20):
    exclude = [x[0] for x in Counter(adata.obs[obskey]).items() if x[1] < num_cut]
    cp = np.array(adata.obs[obskey])
    for item in exclude:
        cp[cp==item] = 'none'
    adata.obs[obskey+'_major'] = cp